"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Default layout for streamlit application
"""
from collections.abc import Callable
from pathlib import Path

import plotly.express as px
import stqdm
import streamlit as st
from pydantic import validate_call

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.recording.folder import FolderRecord
from corl.libraries.functor import Functor
from corl.visualization.episode_artifact_to_tables import EpisodeArtifactTables
from corl.visualization.network_explainability.env_policy_transforms import AgentCheckpoint
from corl.visualization.plotly_animator import FrameAxisBounds
from corl.visualization.streamlit_app.sections.agent_steps import agent_steps_section
from corl.visualization.streamlit_app.sections.platform_and_agent_dones import agent_dones_counts_and_table, platform_dones_counts_and_table
from corl.visualization.streamlit_app.sections.trajectory_animation import trajectory_playback_section
from corl.visualization.streamlit_app.sections.units_conversion import unit_selection_and_table_update
from corl.visualization.streamlit_app.utils import (
    generate_grad_based_feat_importance,
    generate_vf_residuals_dist,
    init_network_explainability_module,
    read_eval_experiment_file,
    st_dataframe_pagination,
    st_directory_picker,
    text_section_policy_network_input,
    use_container_width,
    write_code_block,
)


@st.cache_resource
def episode_artifact_tables(input_path, file_loader):
    """Returns the episode artifact tables"""
    episode_tables = EpisodeArtifactTables(
        input_path=input_path, file_loader=file_loader, status_bar=stqdm.stqdm, print_func=write_code_block
    )
    return episode_tables.tables, episode_tables.evaluation_outcome.episode_artifacts


class DefaultStreamlitPage:
    """Default stremalit app.

    Parameters
    ----------
    file_loader : Callable
        A callable that loads the episode artifacts and test cases. Must return an
        EvaluationOutcome
    """

    def __init__(self, file_loader: Callable = FolderRecord) -> None:
        self.sess_state = st.session_state
        self.file_loader: Callable = file_loader
        self.sess_state.loaded_evaluation = False

        self.tables_dict: dict = {}
        self.episode_artifacts_dict: dict[int, EpisodeArtifact] = {}
        self.obs_units_flattened: dict = {}
        """These are filled out when display_load_episode_artifact_option
        are called"""

        self.full_experiment_path: Path = Path()
        self._root_path: Path = Path()
        self._experiment_dir: str | None = None
        """These are filled out when display_directory_picker is called and a
        valid directory is chosen."""

        self._pages_dict: dict = {}

    def display_directory_picker(self):
        """Displays the directory picker"""
        self._root_path, self._experiment_dir = st_directory_picker(title="Directory Picker")
        if self._experiment_dir is not None and self._root_path.joinpath(self._experiment_dir).exists():
            self.full_experiment_path = self._root_path.joinpath(self._experiment_dir)
            self.sess_state.chosen_experiment_path = True
        else:
            st.warning(
                f""" **Warning: Invalid Selection** - `{self.full_experiment_path}`
            does not contain episode artifacts.
            """
            )

    def display_load_episode_artifact_option(self):
        """Displays the load evaluation button and poopulates the episode artifacts dict
        and tables_dict"""
        if st.button("Load Evaluation"):
            self.sess_state.loaded_evaluation = True
            self.tables_dict, self.episode_artifacts_dict = episode_artifact_tables(self.full_experiment_path, file_loader=self.file_loader)
            self._set_obs_units_dict()

    def _set_obs_units_dict(self):
        _space_table = self.tables_dict["SpaceData"].copy()
        if not _space_table.empty:
            _space_table = _space_table[["attribute_name", "units"]].dropna()
            self.obs_units_flattened = dict(zip(_space_table["attribute_name"], _space_table["units"]))

    def display_config_file(self):
        """Displays the configuration file"""
        if self.sess_state.loaded_evaluation:
            experiment_config, config_file = read_eval_experiment_file(self.full_experiment_path)
            if config_file is not None:
                st.markdown(f"Showing: {self.full_experiment_path.joinpath(config_file)}")
                st.json(experiment_config)

    def display_initial_states_section(self):
        """Displays the initial states table"""
        st.markdown("## Test Cases/ Initial States")
        self._pages_dict = st_dataframe_pagination(tables_dict=self.tables_dict, table_key="InitialStates", pages_dict=self._pages_dict)

    def display_episode_metadata_section(self):
        "Displays the episode metadata section"
        st.markdown("## Episode Metadata")
        self._pages_dict = st_dataframe_pagination(tables_dict=self.tables_dict, table_key="EpisodeMetaData", pages_dict=self._pages_dict)
        fig = px.line(
            data_frame=self.tables_dict["EpisodeMetaData"].sort_values(by=["test_case"]), x="test_case", y="wall_time_sec", markers=True
        )
        st.plotly_chart(fig, use_container_width=use_container_width)

    def display_agent_dones_section(self):
        """Displays the agent done section. When more than one done status is
        selected, shows up as different colored stacked bars"""
        agent_dones_counts_and_table(
            agent_dones_table=self.tables_dict["AgentDoneStatus"], use_container_width=use_container_width, sess_state=self.sess_state
        )

    def display_platform_dones_section(self):
        """Displays the platform done section"""
        platform_dones_counts_and_table(
            platform_dones_table=self.tables_dict["PlatformDoneStates"], use_container_width=use_container_width
        )

    def display_agent_steps_section(self):
        """Displays the agent steps section. A wrapper for `agents_steps_section`"""

        self.sess_state = agent_steps_section(tables_dict=self.tables_dict, sess_state=self.sess_state)

    def display_space_data_table(self):
        """Displays the space data table."""
        space_data_table = self.tables_dict["SpaceData"]
        if not space_data_table.empty:
            st.markdown("## Space Definitions & Units")
            st.dataframe(space_data_table, use_container_width=use_container_width)
            st.markdown(
                """
            **Table**: Provides space (observation and action space) definitions in the original space
            and normalized space."""
            )

    @validate_call
    def display_trajectory_animation_section(self, animator: Functor, frame_axis_bounds: FrameAxisBounds | None = None):
        """
        Displays the Trajectory Animation Section

        Parameters
        ----------
        animator: Functor
            A Functor defining the animator
        frame_axis_bounds: FrameAxisBounds, optional
            Sets the default frame axis bounds for the animation.
        """
        self.sess_state = trajectory_playback_section(
            animator=animator,
            sess_state=self.sess_state,
            episode_artifacts_dict=self.episode_artifacts_dict,
            frame_axis_bounds=frame_axis_bounds,
        )

    def display_policy_network_viz_section(self):
        """Displays the policy network visualization section"""
        st.markdown("# Policy Network Visualizations")
        agent_checkpoints = self._construct_agent_checkpoints()
        env_config_path = self.full_experiment_path.joinpath("env_config.pkl")
        policy_input_map = text_section_policy_network_input(agent_checkpoints=agent_checkpoints)

        clicked_button = st.button("Generate Policy Network Explainability Metrics")

        if clicked_button:
            episode_artifact_files = list(self.full_experiment_path.glob("test_case*/*_episode_artifact.pkl"))

            self.sess_state.eval_net_viz = init_network_explainability_module(
                agent_checkpoints=agent_checkpoints,
                env_config_path=env_config_path,
                policy_network_input_map=policy_input_map,
                episode_artifact_files=episode_artifact_files,
            )
        if self.sess_state.get("eval_net_viz"):
            # #################################################################
            # Displays the Gradient Based Feature Importance
            # #################################################################
            st.markdown("## Policy Network Visualization")
            fig = self.sess_state.eval_net_viz.render_agg_feat_importance_grad_based(normalize=True)
            generate_grad_based_feat_importance(fig)
            # #################################################################
            # Displays the Value Function Residual
            # #################################################################
            fig = self.sess_state.eval_net_viz.render_value_function_residual_dist()
            generate_vf_residuals_dist(fig)

    def _construct_agent_checkpoints(self):
        """Constructs the agent checkpoints list"""
        agent_checkpoints = []
        checkpoints = self.full_experiment_path.joinpath("agent_checkpoints").glob("*")
        for policy_dir in checkpoints:
            agent_name = policy_dir.parts[-1]
            agent_checkpoints.append(AgentCheckpoint(agent_name=agent_name, checkpoint_dir=policy_dir))
        return agent_checkpoints

    def display_units_selector(self):
        """Displays the units selector section"""
        self.tables_dict = unit_selection_and_table_update(tables_dict=self.tables_dict, obs_units_flattened=self.obs_units_flattened)

        # Updates the observation units dictionary
        self._set_obs_units_dict()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Callbacks to do logging of episodes
"""
import collections
import copy
import typing
from abc import ABC, abstractmethod
from datetime import datetime

import flatten_dict
import numpy as np
from ray.rllib import BaseEnv, RolloutWorker
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.typing import AgentID, PolicyID

from corl.environment.base_multi_agent_env import BaseCorlMultiAgentEnv
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.simulators.base_simulator import BaseSimulator


class DefaultEvaluationCallbacks(DefaultCallbacks, ABC):
    """Callbacks to extract data for evaluation reporting.

    This is intended to be used as a subclass of the environment callbacks used by the experiment using cooperative multiple inheritance.
    """

    @abstractmethod
    def platform_serializer(self):
        """
        Function that holds a placeholder for the SerializePlatform object which contains a serialization function.
        The SerializePlatform object is to be instantiated and placed inside this method for future use
        """

    @abstractmethod
    def extract_environment_state(self, env_state: dict):
        """Method that extracts any information in the environment state that needs to be saved"""
        return {}

    def on_episode_start(
        self,  # noqa: PLR6301
        *,
        worker: RolloutWorker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy],
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        """See corl.environment.default_env_rllib_callbacks.EnvironmentDefaultCallbacks.on_episode_start"""

        env = base_env.get_sub_environments()[episode.env_id]
        episode.user_data["initial_state"] = env.observation

        episode.user_data["eval"] = {
            agent_id: {"episode": {"done": {}, "reward": {}, "total_reward": np.nan, "episode_state": {}}, "step": []}
            for agent_id in env.agent_dict
        }

        agent_dones = {}
        for agent_id in env.agent_dict:
            agent_dones[agent_id] = env.agent_dict[agent_id].config.dones

        episode.user_data["test_case"] = env.episode_id
        episode.user_data["env_config"] = worker.config["env_config"]
        episode.user_data["params"] = flatten_dict.flatten(env.local_variable_store, reducer="dot")
        episode.user_data["start_time"] = datetime.now()
        episode.user_data["episode_artifacts_filenames"] = {}
        episode.user_data["observation_units"] = env._observation_units  # noqa: SLF001
        episode.user_data["platform_to_agents"] = env.platform_to_agents
        episode.user_data["agent_to_platforms"] = env.agent_to_platforms
        episode.user_data["done_config"] = EpisodeArtifact.DoneConfig(
            task=env.config.dones.task, world=env.config.dones.world, agent_dones=agent_dones
        )

        step_data: dict[str, dict[str, typing.Any]] = {
            agent_id: {"observation": collections.OrderedDict(), "action": collections.OrderedDict()} for agent_id in env.agent_dict
        }

        for agent_id, obj_pair in env.observation.items():
            step_data[agent_id]["observation"] = obj_pair

        for agent_id in env.agent_dict:
            episode.user_data["eval"][agent_id]["step"].append(step_data[agent_id])

        episode.user_data["steps"] = []
        episode.user_data["walltime_sec"] = []
        episode.user_data["dones"] = {}
        episode.user_data["episode_state"] = {}
        episode.user_data["simulator_info"] = []

        # collect and add full set of parameter values for episode
        parameter_values = {"environment": flatten_dict.flatten(env.local_variable_store, reducer="dot")}
        for agent_id, agent in env.agent_dict.items():
            parameter_values[agent_id] = flatten_dict.flatten(agent.local_variable_store, reducer="dot")
        episode.user_data["parameter_values"] = parameter_values

        # Cooperative multiple inheritance
        super().on_episode_start(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

    def on_episode_step(  # noqa: PLR0912, PLR0915
        self,
        *,
        worker: RolloutWorker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy] | None = None,
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        """See corl.environment.default_env_rllib_callbacks.EnvironmentDefaultCallbacks.on_episode_step"""

        env = base_env.get_sub_environments()[episode.env_id]

        episode.user_data["eval"] = {
            agent_id: {"episode": {"done": {}, "reward": {}, "total_reward": np.nan, "episode_state": {}}, "step": []}
            for agent_id in env.agent_dict
        }

        # Calculate this step's rewards
        map_agent_reward: dict[str, dict[str, float]] = {}
        if env.reward_info is not None:
            for agent_id, obj in env.reward_info.items():
                for name, value in obj.items():
                    if agent_id not in map_agent_reward:
                        map_agent_reward[agent_id] = {}
                    if "__all__" in agent_id:
                        continue
                    map_agent_reward[agent_id].update(self._resolve_sequence_values(name, value))

        # Calculate this step's done
        if env.done_info is not None:
            for agent_id, obj in env.done_info.items():
                for name, value in obj.items():
                    if agent_id not in episode.user_data["dones"]:
                        episode.user_data["dones"][agent_id] = {}
                    if "__all__" in agent_id:
                        continue
                    episode.user_data["dones"][agent_id].update(self._resolve_sequence_values(name, value))

        episode_state: dict = copy.deepcopy(env.state.agent_episode_state)

        # ######################################
        # # Step data - append on each step
        step_data: dict[str, dict[str, typing.Any]] = {
            agent_id: {"observation": collections.OrderedDict(), "action": collections.OrderedDict()} for agent_id in env.agent_dict
        }

        for agent_id, obj_pair in env.observation.items():
            step_data[agent_id]["observation"] = obj_pair

        if env._actions:  # noqa: SLF001
            for agent_id, agent_data in env._actions[-1].items():  # noqa: SLF001
                if isinstance(agent_data, dict):
                    agent_data = flatten_dict.flatten(agent_data)  # noqa: PLW2901
                    for part_name, value in agent_data.items():
                        part_name = ".".join(part_name)  # noqa: PLW2901
                        if isinstance(value, tuple) and len(value) > 0 and isinstance(value[0], dict):
                            # handle RTAModule or otherwise wrapped controllers
                            for elem in value:
                                for sub_part_name, sub_value in elem.items():
                                    step_data[agent_id]["action"][f"{part_name}.{sub_part_name}"] = sub_value
                        else:
                            step_data[agent_id]["action"][part_name] = value
                elif isinstance(agent_data, np.ndarray):
                    step_data[agent_id]["action"] = collections.OrderedDict()
                    for i, val in enumerate(agent_data):
                        step_data[agent_id]["action"][f"action_{i!s}"] = np.array([val])
                else:
                    raise ValueError("Only action types dict and np.ndarray are supported")

                # step_data[agent_id]['action'] = agent_data    # change from corl2.0 merge

        # ##################################################
        # # Append the episode's step information to our agent_data knowledge
        serialized_platforms = [self.platform_serializer().serialize(item) for item in env.state.sim_platforms.values()]

        # # Start our step artifact, the agent map will be filled out next
        step = EpisodeArtifact.Step(agents={}, platforms=serialized_platforms, environment_state=self.extract_environment_state(env.state))

        def get_agent_reward(episode, agent_id):
            try:
                res = episode.agent_rewards[(agent_id, episode.policy_for(agent_id))]
            except KeyError:
                res = 0.0
            return res

        for agent_id, trainable_agent in env.agent_dict.items():
            # If the platform doesn't exists anymore there is nothing to do.
            # # Assumes 1:1 map of agent to platform
            if len([item for item in serialized_platforms if item["name"] == trainable_agent.platform_names[0]]) != 0:
                step.agents[agent_id] = EpisodeArtifact.AgentStep(
                    observations=step_data[agent_id]["observation"],
                    actions=step_data[agent_id]["action"],
                    rewards=map_agent_reward.get(agent_id),
                    total_reward=get_agent_reward(episode, agent_id),
                )

        episode.user_data["steps"].append(step)
        episode.user_data["walltime_sec"].append(datetime.utcnow().timestamp())
        episode.user_data["episode_state"] = episode_state
        episode.user_data["simulator_info"].append(copy.deepcopy(env.state.simulator_info))
        #####################################################
        # Legacy eval setting - overwrite on each step so only last one persists

        for agent_id in env.agent_dict:
            episode.user_data["eval"][agent_id]["step"].append(step_data[agent_id])

        episode.user_data["eval"][agent_id]["episode"]["done"] = episode.user_data["dones"]

        episode.user_data["eval"][agent_id]["episode"]["reward"] = map_agent_reward

        for agent_id in env.agent_dict:
            episode.user_data["eval"][agent_id]["episode"]["total_reward"] = get_agent_reward(episode, agent_id)

        episode.user_data["eval"][agent_id]["episode"]["episode_state"] = episode_state[agent_id]

        # Cooperative multiple inheritance
        super().on_episode_step(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

    def on_postprocess_trajectory(
        self,
        *,
        worker: RolloutWorker,
        episode: EpisodeV2,
        agent_id: AgentID,
        policy_id: PolicyID,
        policies: dict[PolicyID, Policy],
        postprocessed_batch: SampleBatch,
        original_batches: dict[AgentID, tuple[Policy, SampleBatch]],
        **kwargs,
    ) -> None:
        """See corl.environment.default_env_rllib_callbacks.EnvironmentDefaultCallbacks.on_postprocess_trajectory"""

        policy_artifact = {
            agent_id: EpisodeArtifact.PolicyArtifact(
                preprocessor=worker.preprocessors[policy_id],
                filters=worker.filters[policy_id],
            )
        }

        worker_env: BaseCorlMultiAgentEnv | None = worker.env  # type: ignore
        if worker_env is None:
            raise RuntimeError("Worker.env is None, this is a non op")
        if not isinstance(worker_env.simulator, BaseSimulator):
            raise RuntimeError("The simulator attached to the worker does not inherit BaseSimulator")
        frame_rate = worker_env.simulator.frame_rate

        space_definitions = EpisodeArtifact.SpaceDefinitions(
            action_space=worker_env._raw_action_space,  # type: ignore # noqa: SLF001
            normalized_action_space=worker_env.action_space,  # type: ignore
            observation_space=worker_env._raw_observation_space,  # type: ignore  # noqa: SLF001
            normalized_observation_space=worker_env.observation_space,  # type: ignore
        )

        # If we have yet to generate the episode artifact for this episode then do it
        if "episode_artifact" not in episode.user_data:
            episode.user_data["episode_artifact"] = np.array(
                [
                    EpisodeArtifact(
                        test_case=episode.user_data["test_case"],
                        env_config=episode.user_data["env_config"],
                        artifacts_filenames=episode.user_data["episode_artifacts_filenames"],
                        start_time=episode.user_data["start_time"],
                        duration_sec=(datetime.now() - episode.user_data["start_time"]).total_seconds(),
                        steps=episode.user_data["steps"],
                        walltime_sec=episode.user_data["walltime_sec"],
                        dones=episode.user_data["dones"],
                        frame_rate=frame_rate,
                        episode_state=episode.user_data["episode_state"],
                        simulator_info=episode.user_data["simulator_info"],
                        worker_index=worker.env_context.worker_index,
                        params=episode.user_data["params"],
                        parameter_values=episode.user_data["parameter_values"],
                        observation_units=episode.user_data["observation_units"],
                        platform_to_agents=episode.user_data["platform_to_agents"],
                        agent_to_platforms=episode.user_data["agent_to_platforms"],
                        done_config=episode.user_data["done_config"],
                        initial_state=episode.user_data["initial_state"],
                        policy_artifact=policy_artifact,
                        space_definitions=space_definitions,
                        platform_serializer=self.platform_serializer(),
                        algorithm_iteration=self.get_algorithm_iteration(),
                    )
                ]
            )
        else:
            episode.user_data["episode_artifact"][0].policy_artifact.update(policy_artifact)

        if "eval" not in postprocessed_batch:
            postprocessed_batch["eval"] = episode.user_data["episode_artifact"]

        # Cooperative multiple inheritance
        super().on_postprocess_trajectory(
            worker=worker,
            episode=episode,
            agent_id=agent_id,
            policy_id=policy_id,
            policies=policies,
            postprocessed_batch=postprocessed_batch,
            original_batches=original_batches,
            **kwargs,
        )

    def get_algorithm_iteration(self) -> int | None:  # noqa: PLR6301
        """Hook function to return the algorithm iteration.

        Base implementation always returns None, which is proper for usage at evaluation.
        """
        return None

    def _resolve_sequence_values(self, key: str, value: typing.Any) -> dict[str, typing.Any]:
        """Convert a single key with a sequence of values into a sequence of keys each with a single value.

        Parameters
        ----------
        key : str
            Base key for this value
        value : Any
            Value for this key.  This can be a single value, such as a string or number, or a sequence of values.

        Returns
        -------
        Dict[str, Any]
            A dictionary with keys that are the provided key with index values concatenated.  The values are the sequence values at that
            index.
        Example
        >>> self.__resolve_sequence_values('foo', 'bar')
        {'foo': 'bar'}
        >>> self.__resolve_sequence_values('foo', [1, 2, 3])
        {'foo_0': 1, 'foo_1': 2, 'foo_2': 3}
        """
        output: dict[str, typing.Any] = {}
        if isinstance(value, str):
            output[key] = value
        elif isinstance(value, collections.abc.Sequence | np.ndarray):
            if isinstance(value, np.ndarray) and not value.shape:
                output[key] = value.item()
            elif len(value) > 1:
                for index, new_value in enumerate(value):
                    output.update(self._resolve_sequence_values(f"{key}_{index}", new_value))
            elif len(value) == 1:
                output.update(self._resolve_sequence_values(key, value[0]))
            else:
                output[key] = "empty_list"
        elif isinstance(value, collections.abc.Mapping):
            for key2, value2 in value.items():
                output.update(self._resolve_sequence_values(f"{key}.{key2}", value2))
        else:
            output[key] = value
        return output



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Evaluation module
"""



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc
import typing
from collections.abc import MutableMapping

from pydantic import BaseModel, FilePath, ImportString, TypeAdapter, field_validator

from corl.agents.base_agent import AgentParseBase, AgentParseInfo
from corl.libraries.file_path import CorlDirectoryPath
from corl.parsers.agent_and_platform import CorlAgentConfigArgs, CorlPlatformConfigArgs
from corl.parsers.yaml_loader import load_file
from corl.policies.base_policy import BasePolicyValidator


class ExperimentFileParse(BaseModel):
    """[summary]
    experiment_class: The experiment class to run
    config: the configuration to pass to that experiment
    """

    config: FilePath
    compute_platform: str = "local"
    platform_config: list[CorlPlatformConfigArgs]
    agent_config: list[CorlAgentConfigArgs]
    environment: str = "default"
    name: str | None = None
    output: CorlDirectoryPath | None = None
    debug: bool = False


class BaseExperimentValidator(BaseModel):
    """
    Base Validator to subclass for Experiments subclassing BaseExperiment
    """


class BaseExperiment(abc.ABC):
    """
    Experiment provides an anstract class to run specific types of experiments
    this allows users to do specific setup steps or to run some sort of custom training
    loop
    """

    def __init__(self, **kwargs) -> None:
        self.config: BaseExperimentValidator = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseExperimentValidator]:
        """
        Get the validator for this experiment class,
        the kwargs sent to the experiment class will
        be validated using this object and add a self.config
        attr to the experiment class
        """
        return BaseExperimentValidator

    @property
    def get_policy_validator(self) -> type[BasePolicyValidator]:
        """
        Get the policy validator for this experiment class,
        the kwargs sent to the experiment class will
        be validated using this object and add a self.config
        attr to the policy config
        """
        return BasePolicyValidator

    @abc.abstractmethod
    def run_experiment(self, args: ExperimentFileParse):
        """
        Runs the experiment associated with this experiment class

        Arguments:
            args {ExperimentFileParse} -- The args provided by the argparse
                                            in corl.train_rl
        """

    def create_agents(self, platform_configs: list[CorlPlatformConfigArgs], agent_configs: list[CorlAgentConfigArgs]) -> tuple[dict, dict]:
        """Create the requested agents and add them to the environment configuration.

        Parameters
        ----------
        platform_configs : CorlPlatformsConfigArg
            A sequence of platforms
        agent_configs : agent_configs
            A sequence of agents
        """
        platforms = {}
        for platform_config in platform_configs:
            assert platform_config.name not in platforms, "duplicate platforms not allowed"
            platform_configuration = platform_config.config

            platforms[platform_config.name] = platform_configuration

        agents = {}
        for agent_config in agent_configs:
            if not all(platform in platforms for platform in agent_config.platforms):
                raise RuntimeError(
                    f"invalid platform request by agent {agent_config.name}, "
                    f"agent requested platforms {agent_config.platforms}, but available platforms are: "
                    f"{platforms}"
                )

            config = agent_config.config
            policy_config = agent_config.policy

            parsed_agent = AgentParseBase(**config)
            _ = self.get_policy_validator(**policy_config)

            agents[agent_config.name] = AgentParseInfo(
                class_config=parsed_agent, platform_names=agent_config.platforms, policy_config=policy_config
            )

        return agents, platforms

    @staticmethod
    def create_other_platforms(other_platforms_config: typing.Sequence[tuple[str, str]]) -> dict:
        """Create the requested other platforms and add them to the environment configuration.

        Parameters
        ----------
        other_platforms_config : typing.Sequence[typing.Tuple[str, str]]
            A sequence of platforms.  Each platform consists of a name and platform filename.
        """
        other_platforms = {}

        if other_platforms_config:
            for platform_name, platform_file in other_platforms_config:
                platform_config = load_file(platform_file)
                other_platforms[platform_name] = platform_config
        return other_platforms

    @staticmethod
    def _update_config_object(target: dict, path: typing.Sequence[str], arg, func):
        """
        Updates/Modifies a config mapping object.

        target:  Config mapping object
        path:    List of key words identifying where in the target to modify a value
        arg:     Value that will modify or replace a value of target (i.e. 'output')
        func:    Functor that performs action on data in arg (i.e. lambda x : os.path.join(x,'a'))
        """

        def _build_config_dict_entry(local_path, value):
            """
            Given a path, recursively generate a new dict or dict-of-dicts entry for target.
            """
            if len(local_path) == 1:
                return {local_path[0]: value}

            return {local_path.pop(0): _build_config_dict_entry(local_path, value)}

        if isinstance(target, list):
            target.append(_build_config_dict_entry(path, func(arg)))
        elif isinstance(target, MutableMapping):
            embedded_dict = target
            for key in path:
                if key == path[-1]:
                    embedded_dict[key] = func(arg)
                else:
                    embedded_dict = embedded_dict[key]
        else:
            raise TypeError(f"Unsupported type: {type(target).__name__}")

    @staticmethod
    def process_cli_args(config: dict, cli_args: ExperimentFileParse):
        """Modifies/overrides config fields with command line arguments."""


class ExperimentParse(BaseModel):
    """[summary]
    experiment_class: The experiment class to run
    config: the configuration to pass to that experiment
    """

    experiment_class: type[BaseExperiment]
    config: dict[str, typing.Any]

    @field_validator("experiment_class", mode="before")
    @classmethod
    def construct_experiment_class(cls, v):
        """Convert string to callable"""
        if not isinstance(v, type):
            return TypeAdapter(ImportString).validate_python(v)
        return v



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from pydantic import BaseModel

from corl.evaluation.connection.base_epp_update import BaseEppUpdate
from corl.libraries.units import Quantity


class SimpleEppUpdate(BaseModel, BaseEppUpdate):
    separator: str = "."

    updates: dict[str, Quantity | float | str]

    def get_updates(self) -> dict[tuple[str, ...], Quantity | float | str]:
        updates: dict[tuple[str, ...], Quantity | float | str] = {}
        for key, value in self.updates.items():
            updates[tuple(key.split(self.separator))] = value
        return updates



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from enum import Enum

from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.utils.typing import TensorStructType, TensorType

from corl.policies.custom_policy import CustomPolicy, CustomPolicyValidator


class SimplePaddleControllerValidator(CustomPolicyValidator):
    """Validator for SimplePaddleController"""


class SimplePaddleController(CustomPolicy):
    """A simple pong paddle controller policy"""

    def __init__(self, observation_space, action_space, config):
        self.validated_config: SimplePaddleControllerValidator  # type: ignore
        super().__init__(observation_space, action_space, config)

    @staticmethod
    def get_validator() -> type[SimplePaddleControllerValidator]:
        """
        Get the validator for this experiment class,
        the kwargs sent to the experiment class will
        be validated using this object and add a self.config
        attr to the experiment class
        """
        return SimplePaddleControllerValidator

    def custom_compute_actions(
        self,  # noqa: PLR6301
        obs: list[TensorStructType] | TensorStructType,
        platform_obs: dict[str, dict],
        state: list[TensorType] | None = None,
        prev_action: list[TensorStructType] | TensorStructType = None,
        prev_reward: list[TensorStructType] | TensorStructType = None,
        info: dict[str, list] | None = None,
        explore: bool | None = None,
        timestep: int | None = None,
        sim_time: float | None = None,
        agent_id: str | None = None,
        epp_info: dict | None = None,
        episode: EpisodeV2 | None = None,
        **kwargs,
    ) -> TensorType:
        """Computes actions for the current policy.

        Args:
            obs: observations.
            state: List of RNN state input batches, if any.
            prev_action: previous action values.
            prev_reward: previous rewards.
            info: info objects.
            explore: Whether to pick an exploitation or exploration action.
                Set to None (default) for using the value of
                `self.config["explore"]`.
            timestep: The current (sampling) time step.

        Keyword Args:
            kwargs: Forward compatibility placeholder

        Returns:
            actions (TensorType): output actions
        """

        class Action(Enum):
            """Action space for paddle"""

            UP = 0
            NO_MOVE = 1
            DOWN = 2

        control_left_paddle = platform_obs["Obs_Platform_Paddle_Type_Sensor"]["direct_observation"] == 0
        control_right_paddle = platform_obs["Obs_Platform_Paddle_Type_Sensor"]["direct_observation"] == 1
        # ball_position_x = platform_obs['Obs_Ball_Sensor']['direct_observation'][0]
        ball_position_y = platform_obs["Obs_Ball_Sensor"]["direct_observation"][1]
        # ball_velocity_x = platform_obs['Obs_Ball_Sensor']['direct_observation'][2]
        ball_velocity_y = platform_obs["Obs_Ball_Sensor"]["direct_observation"][3]

        left_paddle_position_y = platform_obs["Obs_left_paddle"]["direct_observation"][0]
        left_paddle_height = platform_obs["Obs_left_paddle_size"]["direct_observation"][0]
        right_paddle_position_y = platform_obs["Obs_right_paddle"]["direct_observation"][0]
        right_paddle_height = platform_obs["Obs_right_paddle_size"]["direct_observation"][0]

        if control_left_paddle:
            delta_y = -(ball_position_y + ball_velocity_y - (left_paddle_position_y + left_paddle_height / 2))
        elif control_right_paddle:
            delta_y = -(ball_position_y + ball_velocity_y - (right_paddle_position_y + right_paddle_height / 2))
        else:
            raise RuntimeError("Invalid paddle")

        action: Action
        if delta_y > 0:
            action = Action.UP
        elif delta_y < 0:
            action = Action.DOWN
        else:
            action = Action.NO_MOVE

        return {"PaddleController_Paddle_move": action.value}



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import threading
from typing import TypeVar

import uvicorn
from fastapi import FastAPI, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from pydantic import BaseModel

T = TypeVar("T")


class RestConnectionValidator(BaseModel):
    port: int = 8000
    log_level: str = "info"
    delay_start: bool = False


class BaseRestConnection:
    def __init__(self, **kwargs):
        self._config: RestConnectionValidator  # type: ignore
        super().__init__(**kwargs)

        self.app = FastAPI()

        self.app.get("/")(self.read_root)

        self.app.exception_handler(RequestValidationError)(self.validation_exception_handler)

        self.config = uvicorn.Config(self.app, port=self._config.port, log_level=self._config.log_level, loop="asyncio")
        self.server = uvicorn.Server(self.config)

        self.thread = threading.Thread(target=self.server.run)

        if not self._config.delay_start:
            self.start()

    def start(self):
        self.thread.start()

    def stop(self):
        self.server.should_exit = True
        self.thread.join()

    def validation_exception_handler(self, request: Request, exc: RequestValidationError):  # noqa: PLR6301
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content=jsonable_encoder({"detail": exc.errors(), "body": exc.body}),
        )

    def read_root(self):
        return {"help": f"navigate to {self.config.host}:{self.config.port}/docs"}

    @staticmethod
    def get_validator() -> type[RestConnectionValidator]:
        return RestConnectionValidator

    # def __getstate__(self):
    #     return {}



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Parse the evaluation configuration
"""
import copy
import typing
from collections import abc

# Lowest level of data read from the configuration file
LOW_DATA = int | float | str
LOW_DATA_SEQUENCE = typing.Sequence[LOW_DATA]
# All possibilities for the "value" field in the configuration file
INPUT_VALUE = LOW_DATA | LOW_DATA_SEQUENCE | typing.Mapping[str, int]

# Type of the "randomize" subtree
RANDOMIZE = typing.Mapping[str, LOW_DATA]
# Type for all subtrees under the various variable names
VARIABLE_ELEMENTS = INPUT_VALUE | RANDOMIZE

# Type for each variable within the configuration file
VARIABLE = typing.Mapping[str, VARIABLE_ELEMENTS]
# Type for the collection of all variables.  The code handles deeper recursion, such as Mapping[str, Mapping[str, INPUT_VARIABLE]].
VARIABLE_STRUCTURE = typing.Mapping[str, VARIABLE | typing.Mapping[str, VARIABLE]]

# Type for the collection of all variables after it has been flattened to remove recursion
VARIABLE_STRUCTURE_FLAT = typing.Mapping[str, VARIABLE]
MUTABLE_VARIABLE_STRUCTURE_FLAT = typing.MutableMapping[str, VARIABLE]

# Type for the collection of all variables once a single subtree is selected
MUTABLE_EXTRACTED_VARIABLE_STRUCTURE = typing.MutableMapping[str, VARIABLE_ELEMENTS]
# Type for the collection of all variables once a single subtree -- value -- is selected
EXTRACTED_VARIABLE_STRUCTURE_VALUE = typing.Mapping[str, INPUT_VALUE]
# Type for the collection of all variables once a single subtree -- randomize -- is selected
EXTRACTED_VARIABLE_STRUCTURE_RANDOMIZE = typing.Mapping[str, RANDOMIZE]

# Type for the collection of all variable values after the ranges are resolved
MUTABLE_RESOLVED_VARIABLE_STRUCTURE_VALUE = typing.MutableMapping[str, LOW_DATA_SEQUENCE]
RESOLVED_VARIABLE_STRUCTURE_VALUE = typing.Mapping[str, LOW_DATA_SEQUENCE]


def flatten_config(config: VARIABLE_STRUCTURE, _path: str | None = None) -> MUTABLE_VARIABLE_STRUCTURE_FLAT:
    """Flatten a nested configuration object to a single collection of keys.

    Parameters
    ----------
    config : VARIABLE_STRUCTURE
        Configuration tree as described by `python act3/agents/evaluation/evaluation.py --describe`.

    Returns
    -------
    MUTABLE_VARIABLE_STRUCTURE_FLAT
        Single level configuration tree.  Nested levels are replaced with expanded keys in the format "top_key.sub_key.leaf_key".

    Raises
    ------
    ValueError
        An input key contains the character "."
    TypeError
        The value is not a mapping.

    Notes
    -----
    This function is invertible by `unflatten`.
    """
    output: MUTABLE_VARIABLE_STRUCTURE_FLAT = {}

    for key, value in config.items():
        if "." in key:
            raise ValueError(f'Cannot use "." in key: {key}')
        next_path = key if _path is None else f"{_path}.{key}"

        if not isinstance(value, abc.Mapping):
            raise TypeError(f"Subtree must be a mapping at {next_path}")

        if "value" in value:
            output[next_path] = copy.deepcopy(typing.cast(VARIABLE, value))
        else:
            output.update(flatten_config(typing.cast(typing.Mapping[str, VARIABLE], value), _path=next_path))

    return output


def unflatten(obj: typing.Mapping[str, typing.Any]) -> typing.Mapping[str, typing.Any]:
    """Returned a flattened object to its nested representation

    Parameters
    ----------
    obj : Mapping[str, Any]
        Flattened representation of a nested mapping.  The format of the initial nested mapping is represented by the structure of the keys
        in the format "top_key.sub_key.leaf_key".

    Returns
    -------
    Mapping[str, Any]
        Nested representation of the input mapping.  For example `{'top_key': {'sub_key': {'leaf_key': value}}}`.

    Notes
    -----
    If all values are mappings with the key "value", then this function is invertible by `flatten_config`.
    """
    output: typing.MutableMapping[str, typing.Any] = {}

    for key, value in obj.items():
        key_split = key.split(".")
        suboutput = output
        for subkey in key_split[:-1]:
            if subkey not in suboutput:
                suboutput[subkey] = {}
            suboutput = typing.cast(typing.MutableMapping[str, typing.Any], suboutput[subkey])
        suboutput[key_split[-1]] = value

    return output


def extract_config_element(config: VARIABLE_STRUCTURE_FLAT, element: str = "value") -> MUTABLE_EXTRACTED_VARIABLE_STRUCTURE:
    """Extract the items of a particular element from the dictionary

    Parameters
    ----------
    config : VARIABLE_STRUCTURE_FLAT
        As produced by flatten_config
    element : str, optional
        The element to extract from each dictionary.  The default is "value".

    Returns
    -------
    MUTABLE_EXTRACTED_VARIABLE_STRUCTURE
        Configuration information in the same structure as the input; however, each variable is reduced to only contain the data from its
        value key.

    Examples
    --------
    Input:
    config = {'name': {'value': [0], 'something': 1},
              'group.subkey': {'value': {'min': 1, 'max': 10, 'step': 1}, 'another_key': 3},
              'hello': {'who': 'world'}
              }
    element = 'value'

    Output:
    {'name': [0], 'group.subkey': {'min': 1, 'max': 10, 'step': 1}}

    Raises
    ------
    TypeError
        Values are not mappings.
    """
    output: MUTABLE_EXTRACTED_VARIABLE_STRUCTURE = {}

    for key, data in config.items():
        if not isinstance(data, abc.Mapping):
            raise TypeError(f"Data for {key} is not a mapping")

        if element in data:
            output[key] = data[element]

    return output


def resolve_config_ranges(config: EXTRACTED_VARIABLE_STRUCTURE_VALUE) -> MUTABLE_RESOLVED_VARIABLE_STRUCTURE_VALUE:
    """Resolve single values and min/max/step ranges into explicitly listed value arrays

    Parameters
    ----------
    config : EXTRACTED_VARIABLE_STRUCTURE_VALUE
        Configuration tree as if produced by `extract_config_element` with element `value`.

    Returns
    -------
    MUTABLE_RESOLVED_VARIABLE_STRUCTURE_VALUE
        Configuration tree with the same keys as the input; however, the values are all arrays.  Single values become one-element arrays.
        Min/max/step ranges are expanded into their explicitly listed values.

    Raises
    ------
    ValueError
        Maximum is less than minimum
        Unsupported structure, likely caused by missing "min", "max", or "step" keys, or extra keys provided
    TypeError
        Min/max/step values are not integers
    """

    output: MUTABLE_RESOLVED_VARIABLE_STRUCTURE_VALUE = {}

    for key, data in config.items():
        if isinstance(data, abc.Mapping):
            if set(data.keys()) != {"min", "max", "step"}:
                raise ValueError(f"Unsupported structure at {key}")

            if not all(isinstance(v, int) for v in data.values()):
                raise TypeError(f"Values within {key} are not integers")

            value = data
            if value["min"] > value["max"]:
                raise ValueError(f"Bad range at {key}")

            output[key] = list(range(value["min"], value["max"] + 1, value["step"]))

        elif isinstance(data, str) or not isinstance(data, abc.Sequence):
            # variable is anything that is not a sequence
            output[key] = [data]
        else:
            # variable is already a sequence
            output[key] = list(data)

    return output



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from __future__ import annotations

import dataclasses
import typing
from collections import OrderedDict
from datetime import datetime

import numpy as np
from gymnasium import spaces
from ray.rllib.models.preprocessors import Preprocessor
from ray.rllib.utils.filter import Filter

from corl.dones.done_func_base import DoneStatusCodes
from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer
from corl.libraries.functor import Functor


@dataclasses.dataclass
class EpisodeArtifact:
    """Artifact generated by an episode"""

    @dataclasses.dataclass
    class PolicyArtifact:
        """Postprocessed new observation, value function and value targets
        for one agent.
        """

        preprocessor: Preprocessor
        filters: Filter

    @dataclasses.dataclass
    class AgentStep:
        """Data collected from perspective a an agent at each step"""

        observations: OrderedDict
        actions: OrderedDict
        rewards: dict[str, float] | None
        total_reward: float
        observation_tensors: np.ndarray | None = None
        value_function: float | None = None

    @dataclasses.dataclass
    class Step:
        """Data collected at each timestep"""

        agents: dict[str, EpisodeArtifact.AgentStep]
        """Dictionary of data for each agent
        """

        platforms: list[dict]
        """Serialized platform data
        """

        environment_state: dict[str, typing.Any]
        """Dictionary of environment specific things
        The exact details of this field depends on the environment
        """

    @dataclasses.dataclass
    class SpaceDefinitions:
        """Original and Normalized space definitions"""

        action_space: spaces.Dict
        normalized_action_space: spaces.Dict
        observation_space: spaces.Dict
        normalized_observation_space: spaces.Dict

    @dataclasses.dataclass
    class DoneConfig:
        """Configurations that define dones"""

        task: dict[str, dict | list[Functor]]
        world: list[typing.Any] | None
        agent_dones: dict[str, list]

    test_case: int | None
    env_config: dict[str, typing.Any]
    worker_index: int
    params: dict[str, typing.Any]
    parameter_values: dict[str, typing.Any]  # full parameters set for episode (environment's + agent's local_variable_stores)
    artifacts_filenames: dict[str, str]
    start_time: datetime
    duration_sec: float
    frame_rate: float | None
    steps: list[EpisodeArtifact.Step]
    walltime_sec: list[float]
    dones: dict[str, dict[str, bool]]
    episode_state: dict[str, dict[str, DoneStatusCodes]]
    simulator_info: list[dict[str, str]]
    observation_units: dict[str, typing.Any]
    platform_to_agents: dict[str, typing.Any]
    agent_to_platforms: dict[str, typing.Any]
    done_config: EpisodeArtifact.DoneConfig
    policy_artifact: dict[str, EpisodeArtifact.PolicyArtifact]
    initial_state: OrderedDict
    platform_serializer: PlatformSerializer
    space_definitions: EpisodeArtifact.SpaceDefinitions | None = None
    algorithm_iteration: int | None = None

    def __str__(self) -> str:
        keys = ["test_case", "worker_index", "dones", "algorithm_iteration"]

        info = ", ".join([f"{key}={getattr(self, key)}" for key in keys])
        return f"<{type(self)}[{info}]>"

    def __repr__(self) -> str:
        return str(self)



# mypy: disable-error-code="union-attr"
"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Contains functionality to perform evaluation by iterating over given test cases
"""
import logging
import os
import shutil
from collections import defaultdict
from collections.abc import Iterator
from distutils.dir_util import copy_tree
from pathlib import Path
from threading import Condition
from typing import Any
from weakref import CallableProxyType

import ray.cloudpickle as pickle
import yaml
from ray.rllib.evaluation.rollout_worker import RolloutWorker
from ray.rllib.policy.sample_batch import MultiAgentBatch, SampleBatch, concat_samples
from ray.util.debug import disable_log_once_globally

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.evaluation_outcome import EvaluationOutcome
from corl.evaluation.launchers.base_eval import BaseEvaluator, BaseProcessor, EvalConfig
from corl.evaluation.launchers.evaluate_validator import EvalExperiment
from corl.evaluation.loader.policy_checkpoint import PolicyCheckpoint
from corl.evaluation.recording.i_recorder import IRecord
from corl.evaluation.runners.section_factories.engine.rllib.rllib_trainer import ray_context
from corl.evaluation.runners.section_factories.test_cases.default_strategy import DefaultStrategy
from corl.evaluation.runners.section_factories.test_cases.test_case_manager import TestCaseStrategy


def process_single_episode(worker: RolloutWorker):
    """
    Args:
        worker (_type_): _description_

    Returns:
        _type_: _description_
    """
    batches = [worker.input_reader.next()]
    batch = concat_samples(batches)
    worker.callbacks.on_sample_end(worker=worker, samples=batch)
    worker.output_writer.write(batch)
    return batch


class TestCaseEvaluator(BaseEvaluator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.logger = logging.getLogger(type(self).__name__)
        self._condition = Condition()
        self._running = True

        self._algorithm: CallableProxyType | None = None  # type: ignore

    def __call__(self, config: EvalConfig) -> Iterator[SampleBatch | MultiAgentBatch]:
        evaluator_config = config.experiment
        """Iterates and evaluates all tests cases"""
        # Disable logging because can cause slowdowns
        disable_log_once_globally()

        # The following context must be valid during the running of parallel workers and saving of data
        # Important that this stay here.
        with ray_context(local_mode=evaluator_config.engine.rllib.debug_mode, include_dashboard=False):
            # Do evaluation using ParallelRollouts
            # See Issue https://git.aoc-pathfinder.cloud//ML/act3-rllib-agents/-/issues/1423 for more discussion on this block
            # ParallelRollouts is given a set of test_cases for it to do
            # It will do those test cases, however not all test cases will take the same amount of time
            #
            # When all test cases have been completed a worker will "rollover" and start re-executing previous test case
            # when such a rollover is done in a workers batch it will have a ['eval'[0]['test_case'] value of None
            # This is referred to as an "invalid" test
            # See tabular_parameter_provider.py for rollover logic
            # Setting of \"test_case\" field is done in the EnvironmentCallback
            #
            # Since the rollout will rollover if we don't stop it then it will sample forever
            # The challenge becomes how do we know when all test cases have been processed and we can stop sampling the Parallel rollouts
            # This approach accomplishes this through two maps:
            #  - test_status_scheduled : A map of the test_cases index that have yet to be completed
            #  - test_status_completed : A map of the test_cases that have been completed
            # When a test is done it will be moved from scheduled to completed
            # We know we are done sampling when test_status_scheduled map is empty
            #
            # All that said the while(True) loop this block is based on is scary
            # There is a fear that if a test_case breaks for whatever reason workers
            # will go off re-executing tests for forever.
            # To address this if after a rollout no test are executed we increment a counter.
            # When that counter reaches a threshold then we raise an exception.
            # The exception is thrown because losing a test is potentially a major issue and should be addressed ASAP
            # The threshold is somewhat arbitrary, however it is set to be large, we don't want to quit execution
            # while a valid test is executing.
            self.logger.info("Running Evaluation")

            with self._condition:
                if not self._running:
                    self._algorithm = None
                    return
                self._algorithm = evaluator_config.algorithm

            num_test_cases = evaluator_config.test_case_manager.get_num_test_cases()

            test_status_scheduled: dict[int, bool] = {i: False for i in range(num_test_cases)}
            num_episodes_left = len(test_status_scheduled)
            self.logger.info(f"Running evaluation rollouts, running {num_episodes_left} test cases")
            # Ideally we just call
            # sample_data = synchronous_parallel_sample(worker_set=trainer.workers)
            # however that call will just try to grab a full batch as opposed to just the
            # number of episodes we want, so we will call it's logic manually
            # this may not be required if batch_mode = complete episodes? this might also break
            # other algorithms such as SAC?
            worker_set = self._algorithm.workers
            if worker_set is None:
                raise RuntimeError("No workers available")

            sample_batches: list[SampleBatch | MultiAgentBatch]
            while num_episodes_left > 0:
                if worker_set.num_remote_workers() <= 0:
                    sample_batches = [process_single_episode(worker_set.local_worker())]
                else:
                    remote_workers_to_use = worker_set.healthy_worker_ids()
                    sample_batches = worker_set.foreach_worker_with_id(
                        lambda _id, w: process_single_episode(w),
                        local_worker=False,
                        healthy_only=True,
                        remote_worker_ids=remote_workers_to_use,
                    )
                    if worker_set.num_healthy_remote_workers() <= 0:
                        # There is no point staying in this loop, since we will not be able to
                        # get any new samples if we don't have any healthy remote workers left.
                        break

                with self._condition:
                    if not self._running:
                        return

                    num_episodes_left -= min(len(sample_batches), num_episodes_left)

                    self.logger.info(f"{num_episodes_left} episodes left to run")

                yield concat_samples(sample_batches)

            self.logger.info("Finished running rollouts")
            with self._condition:
                self._algorithm = None

    def reset(self):
        self.logger.warn("Reset not supported. Continuing without resetting...")

    def stop(self):
        with self._condition:
            self._running = False
            worker_set = self._algorithm.workers if self._algorithm is not None else None
            if worker_set is not None:
                worker_set.stop()


class SampleBatchProcessor(BaseProcessor):
    """Converts SampleBatches into records"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.logger = logging.getLogger(type(self).__name__)
        self._running = True

    def __call__(self, config: EvalConfig, results: Any) -> list[IRecord]:
        if not self._running:
            return []

        if not isinstance(
            results,
            SampleBatch | MultiAgentBatch,
        ):
            return []

        return self.process_sample_data(config.experiment, results)

    def stop(self):
        self._running = False

    def process_sample_data(self, config: EvalExperiment, sample_data: SampleBatch | MultiAgentBatch) -> list[IRecord]:
        episode_artifacts: defaultdict[int, list[EpisodeArtifact]] = defaultdict(list)

        if isinstance(sample_data, SampleBatch):
            self.process_sample_batch(config.test_case_manager, sample_data, episode_artifacts)

        elif isinstance(sample_data, MultiAgentBatch):
            for sample_batch in sample_data.policy_batches.values():
                self.process_sample_batch(config.test_case_manager, sample_batch, episode_artifacts)

        # Generate our outcome object
        if isinstance(config.test_case_manager, DefaultStrategy):
            # DefaultStrategy does not plan test cases, so the test case initial conditions need to be manually
            # added to the test case list after the episode to save the outcome
            for episode_artifact in episode_artifacts.values():
                for artifact in episode_artifact:
                    init_cond = {f"environment.{state}": value for state, value in artifact.params.items()}
                    init_cond["test_case"] = artifact.test_case
                    config.test_case_manager.test_cases.append(init_cond)

        outcome = EvaluationOutcome(config.test_case_manager.get_test_cases(), episode_artifacts)

        records: list[IRecord] = []
        for recorder in config.recorders:
            record = recorder.resolve()
            record.save(outcome)
            records.append(record)

        return records

    def process_sample_batch(
        self, test_case_manager: TestCaseStrategy, sample_batch: SampleBatch, episode_artifacts: defaultdict[int, list[EpisodeArtifact]]
    ):
        """Iterate over the completed batches and update our scheduled/completed states accordingly"""
        episode_artifact: EpisodeArtifact
        for episode_artifact in sample_batch["eval"]:
            if episode_artifact.test_case is not None:
                # Check to see if new test case for this item... If new process it results
                test_case_id = test_case_manager.get_test_case_index(episode_artifact.test_case)
                for key in episode_artifact.episode_state:
                    self.logger.debug(f"{key}:{episode_artifact.test_case}:{episode_artifact.episode_state[key]}")
                # Save off the artifacts and update scheduled state
                episode_artifacts[test_case_id].append(episode_artifact)


class ConfigSaver(BaseProcessor):
    """Saves env_config and checkpoints with records"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._evaluation_output_paths: list[Path] = []  # type: ignore
        self._running = True

    def stop(self):
        self._running = False

    def __call__(self, config: EvalConfig, results: Any):
        if not self._running:
            return

        if not isinstance(results, list):
            results = [results]
        if not all(isinstance(r, IRecord) for r in results):
            return

        result: IRecord
        for result in results:
            if result.absolute_path not in self._evaluation_output_paths:
                self.save_config(config, result.absolute_path)
                self._evaluation_output_paths.append(result.absolute_path)

    @staticmethod
    def save_config(config: EvalConfig, directory: Path):
        """Save env_config and checkpoints to output_path"""
        os.makedirs(directory, exist_ok=True)

        # Copies the configuration file
        config_filename = directory / config.path.name if config.path is not None else directory / "config.yml"

        with open(config_filename, "w") as file_obj:
            yaml.dump(config.raw_config, file_obj)

        env_config = config.experiment.experiment.env_config
        env_config_filename = directory / "env_config.pkl"
        with open(env_config_filename, "wb") as file_obj:
            pickle.dump(env_config, file_obj)

        # Copies the checkpoints for each agent
        checkpoint_dir = directory / "agent_checkpoints"
        os.makedirs(checkpoint_dir)

        for agent_config in config.experiment.teams.agent_config:
            if isinstance(agent_config.agent_loader, PolicyCheckpoint):
                checkpoint_filename = agent_config.agent_loader.checkpoint_filename
                if checkpoint_filename is not None:
                    agent_checkpoint_dir = checkpoint_dir / agent_config.name
                    os.makedirs(agent_checkpoint_dir)
                    if os.path.isdir(checkpoint_filename):
                        copy_tree(str(checkpoint_filename), str(agent_checkpoint_dir))
                    else:
                        shutil.copy(checkpoint_filename, agent_checkpoint_dir)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Done condition for GymnasiumSimulator
"""
from corl.dones.done_func_base import DoneFuncBase


class GymnasiumDone(DoneFuncBase):
    """
    A done functor that simply mirrors the done condition coming from the sim state
    this only works with the GymnasiumSimulator
    """

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space,
        observation_units,
    ):
        return next_state.terminated[self.config.platform_name] or next_state.truncated[self.config.platform_name]



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Abstract class to load agents
"""
import re
from abc import abstractmethod

from ray.rllib.algorithms import Algorithm
from ray.rllib.utils.typing import ModelWeights


class IAgentLoader:
    """Abstract class to load agents"""

    @abstractmethod
    def apply_to_algorithm(self, algorithm: Algorithm, policy_to_apply: str) -> None:
        """Applies the agent to a RLLIB algorithm

        Arguments:
            algorithm {Algorithm} -- Algorithm to apply agent to
            policy_to_apply {str} -- Policy in the algorithm to apply weights to
        """

    @property
    def env_config(self) -> dict | None:
        """Returns the env config used for generating this checkpoint
        Failing to return an env_config indicates that this is non-trainable

        Returns
        -------
        env_config
        """
        return None

    @property
    def agent_id(self) -> str | None:
        """Returns the agent_id (during training) of the checkpoint
        Failing to return an agent_id indicates that this is non-trainable
        Returns
        -------
        agent_id: typing.Optional[str]
        """
        return None

    @staticmethod
    def _apply_policy_to_weights(algorithm: Algorithm, weights: ModelWeights, policy_to_apply: str):
        """Apply a set of weights intended for a given policy to an algorithm

        This method will attempt to resolve weight naming mismatches between pytorch and tensorflow
        Millage may vary

        Arguments:
            algorithm {Algorithm} -- Algorithm to apply weights to
            weights {ModelWeights} -- Given weights to insert into algorithm
            policy_to_apply {str} -- Policy_id in algorithm to apply weights to

        Raises:
            RuntimeError: Thrown if the given weight doesn't match a weight in the architecture
        """

        # Determine if the architecture is expecting a policy name in the weights
        first_expected_weight = next(iter(algorithm.workers.local_worker().get_weights()[policy_to_apply].keys()))  # type: ignore
        arch_policy_id_prefix = policy_to_apply == first_expected_weight[: len(policy_to_apply)]

        # determine the architecture's delimiter
        arch_delimeter = None
        if "/" in first_expected_weight:
            arch_delimeter = "/"
        elif "." in first_expected_weight:
            arch_delimeter = "."

        # Sanity check
        if arch_policy_id_prefix and arch_delimeter is None:
            raise RuntimeError("The architecture expects a policy, but there is not a recognized delimiter")

        # Iterate over all weights
        weights_renamed = ModelWeights()
        for key, value in weights.items():
            # Tensorflow/pytorch compatibility
            updated_key = key

            # Look for a prefix of either {policy_id} or a <team><num>
            # Due to pytorch/tensorflow the prefix may or may not exists
            # This code assumes if there is a prefix it will be one of two things:
            #  - "{policy_id}" as set by Training Algorithm (LP)
            #  - <team_name><num> as set by tensor flow checkpoints (I think?)
            #  - Nothing as set by pytorch
            #
            # If the prefix is found remove it and the following character (delimiter)
            # A prefix will be potentially be added back in the future according to architecture
            prefix_to_remove_regex = rf"^(?:{policy_to_apply}|(?:(?:red|blue|green|grey)\d+))"

            m = re.search(prefix_to_remove_regex, updated_key)
            if m is not None:
                updated_key = updated_key[len(m.group(0)) + 1 :]

            # If we know the architecture is expecting a delimenter then convert the
            # given weights key to that delimiter
            if arch_delimeter is not None:
                if arch_delimeter == ".":
                    if "/" in updated_key:
                        updated_key = updated_key.replace("/", ".")
                elif arch_delimeter == "/":
                    if "." in updated_key:
                        updated_key = updated_key.replace(".", "/")
                else:
                    raise RuntimeError(f"Unknown delimiter encountered: {arch_delimeter}")

            # If expecting the policy id append
            if arch_policy_id_prefix:
                updated_key = f"{policy_to_apply}{arch_delimeter}{updated_key}"

            assert (
                updated_key in algorithm.workers.local_worker().get_weights()[policy_to_apply]  # type: ignore
            ), f"The {updated_key} weight does not match a known weight in the architecture!"

            weights_renamed[updated_key] = value

        assert (
            weights_renamed.keys() == algorithm.workers.local_worker().get_weights()[policy_to_apply].keys()  # type: ignore
        ), "The model used by rllib config and the model that weights are being loaded from don't match in architecture!"

        algorithm.workers.local_worker().set_weights({policy_to_apply: weights_renamed})  # type: ignore



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

Integration module provides abstraction for integration to enable RL environment connect to
simulation environments and/or real environments. The concept is that if we build either
simulation or real environments from this common interface we can transition between them.

- For example if there is an Sim1StickController and a Sim2StickController the interface
between them is derived from BaseStickController and thus to the policy the interface will
be the same.

The base classes have properties which determine what the deriving classes must
adhere to in order to comply with the intention of the base class.

- For example the BasePlatform has a position property which has a position_properties
(a MultiBoxProp object) that determines the ranges and types of the position property for
the BasePlatform.

To differentiate from the agent properties, base integration nomenclature uses control and
measurements for the base parts, in contrast to the nomenclature of the agent which is usually
action and observation. In this sense actions are made up of controls, and observations are
made up of measurements.
"""
from __future__ import annotations

import abc
import numbers
import typing
import warnings
from operator import attrgetter
from typing import Annotated

import tree
from pydantic import BaseModel, BeforeValidator, ImportString, model_validator

import corl.simulators.base_properties as base_props
from corl.libraries.nan_check import nan_check_result
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import Prop
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_simulator_state import BaseSimulatorState


class BasePlatformPartValidator(BaseModel):
    """
    name: the optional name for this part, the class name
    will be used otherwise
    """

    part_class: ImportString
    name: str | None = None
    initial_validity: bool = True
    properties: dict | None = {}

    @model_validator(mode="after")
    def check_name(self):
        """Check if agent subclass AgentBase"""
        if self.name is None:
            v = PluginLibrary.FindGroup(self.part_class)
            self.name = v
        return self


class BasePlatformPart(abc.ABC):
    """
    BasePlatformPart abstract class for the classes that will be part of the BasePlatform.
    This includes controls and sensors
    """

    def __init__(self, parent_platform, config, property_class) -> None:
        config["part_class"] = self.__class__
        self.config = self.get_validator()(**config)
        self._property_class = property_class
        self._properties = property_class(**self.config.properties)
        self._parent_platform = parent_platform
        self._valid = self.config.initial_validity

    @property
    def valid(self) -> bool:
        """
        Signifies if measurements from this part should be trusted as having
        data that is valid or important to training

        Returns
        -------
        bool
            The current state of the part being valid or not
        """
        return self._valid

    def set_valid(self) -> None:
        """
        Signifies that this part should transition to being valid
        """
        self._valid = True

    def set_invalid(self) -> None:
        """
        Signifies that this part should transition to being invalid
        """
        self._valid = False

    @property
    def name(self) -> str | None:
        """
        The name for this platform part

        Returns
        -------
        str
            The name for this platform part
        """
        return self.config.name

    @property
    def parent_platform(self) -> BasePlatform:  # type: ignore  # noqa: F821
        """
        The parent platform this platform part is attached to

        Returns
        -------
        BasePlatform
            The parent platform this platform part is attached to
        """
        return self._parent_platform

    @staticmethod
    def get_validator() -> type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return BasePlatformPartValidator

    @classmethod
    def embed_properties(cls, property_class: type[Prop]) -> type[BasePlatformPart]:
        """Embed the properties in the class definition."""

        class DynamicPlatformPart(cls):  # type: ignore[valid-type,misc]
            def __init__(self, parent_platform, config) -> None:
                super().__init__(parent_platform=parent_platform, config=config, property_class=property_class)

        DynamicPlatformPart.__doc__ = cls.__doc__
        DynamicPlatformPart.__name__ += f":{cls.__name__}:{property_class.__name__}"
        DynamicPlatformPart.__qualname__ += f":{cls.__name__}:{property_class.__name__}"
        DynamicPlatformPart.embedded_properties = property_class

        return DynamicPlatformPart


nested_control_dict = typing.Mapping[tuple[str, ...], Quantity]
ControlType = nested_control_dict | Quantity


class BaseController(BasePlatformPart, abc.ABC):
    """
    BaseController base abstraction for a controller. A controller is used to move a platform with action commands.
    The actions are usually changing the desired rates or applied forces to the platform.
    """

    def __init__(self, parent_platform, config, property_class) -> None:
        super().__init__(parent_platform, config, property_class)
        self._prop_space = self.control_properties.create_space()

    @property
    def control_properties(self) -> Prop:
        """
        The properties of the control given to the apply_control function

        Returns
        -------
        Prop
            The properties of the control given tot he apply_control function
        """
        return self._properties

    def validate_control(self, control: ControlType) -> None:
        """
        The generic method to validate a control for this controller.

        Parameters
        ----------
        control
            The control to be validated
        """
        if not self._prop_space.contains(tree.map_structure(lambda x: x.m, control)):
            raise ValueError(f"{type(self).__name__} control {control} not in space {self._prop_space} values")

    @abc.abstractmethod
    def apply_control(self, control: ControlType) -> None:
        """
        The generic method to apply the control for this controller.

        Parameters
        ----------
        control
            The control to be executed by the controller
        """

    @abc.abstractmethod
    def get_applied_control(self) -> ControlType | numbers.Number:
        """
        Get the previously applied control that was given to the apply_control function
        Returns
        -------
        previously applied control that was given to the apply_control function
        """

    def get_validated_applied_control(self) -> ControlType:
        """
        Get the previously applied control with nan check
        Returns
        -------
        previously applied control that was given to the apply_control function
        """
        return nan_check_result(self.get_applied_control())


class BaseSensor(BasePlatformPart, abc.ABC):
    """
    BaseSensor base abstraction for a sensor. A sensor is a attached to a platform
    and provides information about the environment.
    """

    def __init__(self, parent_platform, config, property_class) -> None:
        super().__init__(parent_platform=parent_platform, config=config, property_class=property_class)
        self._last_measurement: Quantity | None = None

    @property
    def measurement_properties(self) -> Prop:
        """
        The properties of the object returned by the get_measurement function

        Returns
        -------
        Prop
            The properties of the measurement returned by the get_measurement function
        """
        return self._properties

    @abc.abstractmethod
    def _calculate_measurement(self, state: BaseSimulatorState) -> Quantity:
        """
        The generic method to get calculate measurements from this sensor. This is used to calculate
         the measurement to be returned by
        the get_measurement function. This allows caching the measurement to avoid re-calculation

        Parameters
        ----------
        state: BaseSimulatorState
            The current state of the simulation used to obtain the measurement

        Returns
        -------
        typing.Union[np.ndarray, typing.Tuple, typing.Dict]
            The measurements from this sensor
        """

    def calculate_and_cache_measurement(self, state: BaseSimulatorState):
        """
        Calculates the measurement and caches the result in the _last_measurement variable

        Parameters
        ----------
        state: BaseSimulatorState
            The current state of the simulator used to obtain the measurement
        """
        measurement = self._calculate_measurement(state)
        try:
            nan_check_result(measurement, True)
            self._last_measurement = measurement
        except ValueError:
            warnings.warn(
                "The mover has produced a state that is invalid - NaNs --- Code is going to set broken/damaged - Reuse last state"
            )
            # raise ValueError(f"Error calculating Measurement in {self.__class__}\n" f"Measurement: {self._last_measurement}\n") from err
        if self._last_measurement is None:
            raise ValueError("Measurement is None")

    def get_measurement(self) -> Quantity:
        """
        The generic method to get measurements from this sensor.

        Returns
        -------
        typing.Union[np.ndarray, typing.Tuple, typing.Dict]
            The measurements from this sensor
        """
        if self._last_measurement is None:
            raise ValueError(f"Measurement is None - may also want to check operable states - ({type(self)})")
        return self._last_measurement


class CommandWithArgs(BaseModel):  # type: ignore[no-redef]
    """Model for a command with its arguments."""

    command: str
    args: list[typing.Any] = []
    kwargs: dict[str, typing.Any] = {}


class NoOpControllerValidator(BasePlatformPartValidator):
    """Validator for NoOpController

    platform_init_commands: Allow the user to specify platform initialization commands.  The primary purpose of this is to allow
    verification testing between different simulators under different initial conditions without the complexity of actions being
    sent to those simulators.

    Dot notation is supported, so `foo.bar.baz` will call the method `self.parent_platform.foo.bar.baz()`.

    """

    platform_init_commands: list[Annotated[CommandWithArgs, BeforeValidator(lambda v: {"command": v} if isinstance(v, str) else v)]] = []


class NoOpController(BaseController):
    """
    NoOpController controller define by empty prop can be use when simulator is
    creating actions for an agent.
    """

    def __init__(self, parent_platform, config, property_class) -> None:
        self.config: NoOpControllerValidator
        super().__init__(parent_platform=parent_platform, config=config, property_class=property_class)
        self.new_control = corl_get_ureg().Quantity(0, "dimensionless")

        for command_data in self.config.platform_init_commands:
            func = attrgetter(command_data.command)
            func(self.parent_platform)(*command_data.args, **command_data.kwargs)

    @staticmethod
    def get_validator() -> type[NoOpControllerValidator]:
        """Validator for NoOpController"""
        return NoOpControllerValidator

    def apply_control(self, control) -> None:
        """
        The generic method to apply the control for this controller.

        Parameters
        ----------
        control
            The control to be executed by the controller
        """
        self.new_control = control

    def get_applied_control(self):
        return self.new_control


PluginLibrary.AddClassToGroup(NoOpController.embed_properties(base_props.NoOpProp), "Controller_NoOp", {})


class BaseTimeSensor(BaseSensor):
    """Base type for a time sensor

    Arguments:
        BaseSensor -- The base class type for the sensor
    """

    def __init__(self, parent_platform, config: dict | None = None):
        super().__init__(parent_platform, config, base_props.TimeProp)



from ray.rllib.algorithms.algorithm import Algorithm

from corl.environment.multi_agent_env import ACT3MultiAgentEnv


def reset_epp(algorithm: Algorithm):
    """Resets all episode parameter providers"""
    worker_set = algorithm.workers
    if worker_set is None:
        raise RuntimeError("No workers available")

    def reset_env_epp(env) -> list:
        if isinstance(env, ACT3MultiAgentEnv):
            for epp in env.config.epp_registry.values():
                epp.reset()
        return []

    worker_set.foreach_env(reset_env_epp)


def cleanup_algorithm(algorithm: Algorithm):
    """Explicitly deletes all environments in algorithm to ensure that they can be garbage collected"""
    algorithm.stop()

    def try_shutdown(env):
        if isinstance(env, ACT3MultiAgentEnv):
            env.simulator.shutdown()

    if algorithm.workers is not None:
        try_shutdown(algorithm.workers.local_worker().env)
        del algorithm.workers.local_worker().env

        try_shutdown(algorithm.workers.local_worker().config.env)
        del algorithm.workers.local_worker().config.env

        try:
            for env in algorithm.workers.local_worker().async_env.envs:
                try_shutdown(env)

            del algorithm.workers.local_worker().async_env.envs[:]
        except AttributeError:
            pass

        try:
            for env_state in algorithm.workers.local_worker().async_env.env_states:
                try_shutdown(env_state.env)
                del env_state.env
        except AttributeError:
            pass

        try:
            try_shutdown(algorithm.workers.local_worker().async_env._unwrapped_env)  # noqa: SLF001
            del algorithm.workers.local_worker().async_env._unwrapped_env  # noqa: SLF001
        except AttributeError:
            pass



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from collections import OrderedDict
from functools import cached_property

import numpy as np
from gymnasium.spaces import Dict
from pydantic import BaseModel, model_validator

from corl.glues.base_glue import BaseAgentPlatformGlue, BaseAgentPlatformGlueValidator
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import Quantity, corl_get_ureg, corl_quantity


class LimitConfigValidator(BaseModel):
    """
    minimum: defines the min the obs of the glue may produce
    maximum: defines the max the obs of the glue may produce
    unit: defines the unit for the limits
    clip: defines if the glue should clip the obs values to between min/max
    """

    minimum: Quantity
    maximum: Quantity
    unit: str = "dimensionless"

    @model_validator(mode="before")
    @classmethod
    def check_units(cls, values):
        """
        This function will attmempt to align all of the
        units between the different units that can be
        specified in this validator
        """
        # if the user does not pass a quantity, we attempt to either make a quantity out of it or let it blow up
        if isinstance(values["minimum"], int | float):
            min_val = corl_quantity()(values["minimum"], values.get("unit", "dimensionless"))
        else:
            min_val = values["minimum"]

        if isinstance(values["maximum"], int | float):
            max_val = corl_quantity()(values["maximum"], values.get("unit", "dimensionless"))
        else:
            max_val = values["maximum"]

        # if the limit validator is dimensionless, we will assume min and max are the same unit
        if values.get("unit", "dimensionless") == "dimensionless":
            if min_val.u != max_val.u:
                raise ValueError(f"LimitConfigValidator did not receive a unit, but {min_val=} does not have the same unitas {max_val=}")

            values["unit"] = str(min_val.u)
        # if the limit validator has a unit, we will convert min and max to that
        min_is_comp = min_val.is_compatible_with(values["unit"])
        max_is_comp = max_val.is_compatible_with(values["unit"])

        if not (min_is_comp and max_is_comp):
            raise ValueError(
                f"LimitConfigValidator was told it's limits were dimension {values['unit']}",
                f"but your unit registry does not believe {min_val=} or {max_val=} are compatible"
                f"with this unit type.  {min_is_comp=}, {max_is_comp=}",
            )

        values["minimum"] = min_val.to(values["unit"])
        values["maximum"] = max_val.to(values["unit"])
        return values


class TargetValueValidator(BaseAgentPlatformGlueValidator):
    """
    target_value: the Value this glue should produce this episode
    unit: the unit of the value produced
    """

    target_value: Quantity
    limit: LimitConfigValidator

    @model_validator(mode="after")
    def check_tv_units(self):
        """
        will convert the target value units to the units of the limit
        """
        try:
            self.target_value = self.target_value.to(self.limit.unit)
        except Exception as exc:  # noqa
            raise ValueError(f"TargetValueValidator Tried to convert {self.target_value=} to {self.limit.unit=}, but could not") from exc
        return self


class TargetValue(BaseAgentPlatformGlue):
    """
    Glue class for an observation of some constant value
    """

    class Fields:
        """
        field data
        """

        TARGET_VALUE = "target_value"

    def __init__(self, **kwargs) -> None:
        self.config: TargetValueValidator
        super().__init__(**kwargs)

        self._uname = f"{self.config.name}TargetValue"

        self._target_value = OrderedDict()
        tmp = corl_get_ureg().Quantity(np.asarray([self.config.target_value.m], np.float32), self.config.limit.unit)
        assert isinstance(self.observation_space, Dict)
        tmp = EnvSpaceUtil.clip_space_sample_to_space(tmp, self.observation_space[self.Fields.TARGET_VALUE])
        self._target_value[self.Fields.TARGET_VALUE] = tmp

    @staticmethod
    def get_validator() -> type[TargetValueValidator]:
        return TargetValueValidator

    def get_unique_name(self):
        """Provides a unique name of the glue to differentiate it from other glues."""
        return self._uname

    def invalid_value(self) -> OrderedDict:
        """When invalid return a value of 0

        TODO: this may need to be self.min in the case that the minimum is larger than 0 (i.e. a harddeck)

        Returns:
            OrderedDict -- Dictionary with <FIELD> entry containing 1D array
        """
        return self._target_value

    @cached_property
    def observation_prop(self):
        # DictProp, BoxProp
        return DictProp(
            spaces={
                self.Fields.TARGET_VALUE: BoxProp(
                    low=[self.config.limit.minimum.m_as(self.config.limit.unit)],
                    high=[self.config.limit.maximum.m_as(self.config.limit.unit)],
                    dtype=np.dtype(np.float32),
                    unit=self.config.limit.unit,
                )
            }
        )

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> OrderedDict:
        """Constructs a portion of the observation space

        Arguments:
            platforms {List[BasePlatform]} -- List of current platforms.
                Guaranteed to be length of 1 and set to self.obs_agent

        Returns:
            OrderedDict -- Dictionary containing observation in <FIELD> key
        """

        return self._target_value



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

Normalizer
"""
import abc
import copy
from collections import OrderedDict

import gymnasium.spaces
import numpy as np
from pydantic import BaseModel, field_validator

from corl.libraries.env_space_util import EnvSpaceUtil


class Normalizer(abc.ABC):
    """Class defining normalization of space and samples"""

    def __init__(self, **kwargs) -> None:
        self.config = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator():
        """
        get validator for this Normalizer
        """

    @abc.abstractmethod
    def normalize_space(self, space: gymnasium.spaces.Space) -> gymnasium.spaces.Space:
        """
        normalize_space

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space to be normalized
        Returns:
            gymnasium.spaces.Space
                Normalized space
        """

    @abc.abstractmethod
    def normalize_sample(self, space: gymnasium.spaces.Space, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        normalize_sample

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space
            sample: EnvSpaceUtil.sample_type
                unnormalized sample space to be normalized
        Returns:
            EnvSpaceUtil.sample_type
                Normalized sample
        """

    @abc.abstractmethod
    def unnormalize_sample(self, space: gymnasium.spaces.Space, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        unnormalize_sample

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space
            sample: EnvSpaceUtil.sample_type
                normalized sample space to be unnormalized
        Returns:
            EnvSpaceUtil.sample_type
                Unnormalized sample
        """


class LinearNormalizerValidator(BaseModel):
    """LinearNormalizerValidator

    Parameters:
        minimum: float
            minimum normalized value
        maximum: float
            maximum normalized value
    """

    minimum: float = -1.0
    maximum: float = 1.0


class LinearNormalizer(Normalizer):
    """Class defining linear normalization of space and samples"""

    def __init__(self, **kwargs) -> None:
        self.config: LinearNormalizerValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[LinearNormalizerValidator]:
        """get validator for this LinearNormalizer

        Returns:
            LinearNormalizerValidator
        """
        return LinearNormalizerValidator

    def normalize_space(self, space: gymnasium.spaces.Space) -> gymnasium.spaces.Space:
        """
        normalize_space

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space to be normalized
        Returns:
            gymnasium.spaces.Space
                Normalized space between minimum and maximum
        """

        return EnvSpaceUtil.normalize_space(space=space, out_min=self.config.minimum, out_max=self.config.maximum)

    def normalize_sample(self, space: gymnasium.spaces.Space, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        normalize_sample

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space
            sample: EnvSpaceUtil.sample_type
                unnormalized sample space to be normalized
        Returns:
            EnvSpaceUtil.sample_type
                Normalized sample between min and max output value
        """
        return EnvSpaceUtil.scale_sample_from_space(
            space=space, space_sample=sample, out_min=self.config.minimum, out_max=self.config.maximum
        )

    def unnormalize_sample(self, space: gymnasium.spaces.Space, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        unnormalize_sample

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space
            sample: EnvSpaceUtil.sample_type
                normalized sample space from min to max vale be unnormalized
        Returns:
            EnvSpaceUtil.sample_type
                Unnormalized sample
        """
        return EnvSpaceUtil.unscale_sample_from_space(
            space=space, space_sample=sample, out_min=self.config.minimum, out_max=self.config.maximum
        )


class StandardNormalNormalizerValidator(BaseModel):
    """StandardNormalNormalizerValidator

    Parameters:
        mu: float
            mean of distribution
        sigma: float
            standard devivation
    """

    mu: list[float] = [0.0]
    sigma: list[float] = [1.0]

    @field_validator("mu", "sigma", mode="before")
    @classmethod
    def check_iterable(cls, v):
        """
        Check if mu, sigma are iterable.
        """
        if isinstance(v, float):
            v = [v]
        return v


class StandardNormalNormalizer(Normalizer):
    """Class defining standard norm normalization of space and samples"""

    def __init__(self, **kwargs) -> None:
        self.config: StandardNormalNormalizerValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[StandardNormalNormalizerValidator]:
        """get validator for this StandardNormalNormalizer

        Returns:
            StandardNormalNormalizerValidator
        """
        return StandardNormalNormalizerValidator

    def normalize_space(self, space: gymnasium.spaces.Space) -> gymnasium.spaces.Space:
        """
        normalize_space

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space to be normalized
        Returns:
            gymnasium.spaces.Space
                Normalized space between minimum and maximum
        """
        return EnvSpaceUtil.iterate_over_space(
            func=StandardNormalNormalizer.__standard_normalize_space,
            space=space,
            mu=self.config.mu,
            sigma=self.config.sigma,
        )

    def normalize_sample(self, space: gymnasium.spaces.Space, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """normalize_sample

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space
            sample: EnvSpaceUtil.sample_type
                unnormalized sample space to be normalized

        Returns:
            EnvSpaceUtil.sample_type
                Normalized sample between min and max output value
        """
        return EnvSpaceUtil.iterate_over_space_with_sample(
            func=StandardNormalNormalizer.standard_normalize_sample, space=space, sample=sample, mu=self.config.mu, sigma=self.config.sigma
        )

    def unnormalize_sample(self, space: gymnasium.spaces.Space, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        unnormalize_sample

        Parameters:
            space: gymnasium.spaces.Space
                unnormalized space
            sample: EnvSpaceUtil.sample_type
                normalized sample space from min to max vale be unnormalized
        Returns:
            EnvSpaceUtil.sample_type
                Unnormalized sample
        """
        tmp = EnvSpaceUtil.iterate_over_space_likes(
            func=StandardNormalNormalizer.unnormalize_sample_from_mu_sigma,
            space_likes=(space, sample),
            mu=self.config.mu,
            sigma=self.config.sigma,
            return_space=False,
        )
        if not isinstance(tmp, gymnasium.spaces.Space):
            return tmp
        raise RuntimeError("bug spat, should never happen")

    @staticmethod
    def __standard_normalize_space(
        space_likes: tuple[gymnasium.spaces.Space, EnvSpaceUtil.sample_type],
        mu: float = 0.0,
        sigma: float = 1.0,
    ) -> gymnasium.spaces.Space:
        """
        Normalizes a given gymnasium box using the provided mu and sigma.

        Parameters
        ----------
        space_likes: typing.Tuple[gymnasium.spaces.Space]
            The gymnasium space to turn all boxes into the scaled space.
        mu: float = 0.0
            Mu for normalization.
        sigma: float = 1.0
            Sigma for normalization.

        Returns
        -------
        gymnasium.spaces.Space:
            The new gymnasium spaces where all boxes have had their bounds changed.
        """
        space_arg = space_likes[0]
        if isinstance(space_arg, gymnasium.spaces.Box):
            low = np.divide(np.subtract(space_arg.low, mu), sigma)
            high = np.divide(np.subtract(space_arg.high, mu), sigma)
            return gymnasium.spaces.Box(low=low, high=high, shape=space_arg.shape, dtype=np.float32)
        return copy.deepcopy(space_arg)

    @staticmethod
    def standard_normalize_sample(
        space_likes: tuple[gymnasium.spaces.Space, OrderedDict | (dict | (tuple | (np.ndarray | list)))],
        mu: float = 0.0,
        sigma: float = 1,
    ) -> OrderedDict | (dict | (tuple | (np.ndarray | list))):
        """
        This normalizes a sample from a box space using the mu and sigma arguments.

        Parameters
        ----------
        space_likes: typing.Tuple[gymnasium.spaces.Space, sample_type]
            The first element is the gymnasium space.
            The second element is the sample of this space to scale.
        mu: float
            The mu used for normalizing the sample.
        sigma: float
            The sigma used for normalizing the sample.

        Returns
        -------
        typing.Union[OrderedDict, dict, tuple, np.ndarray, list]:
            The normalized sample.
        """
        (space_arg, space_sample_arg) = space_likes
        if isinstance(space_arg, gymnasium.spaces.Box):
            assert isinstance(space_sample_arg, np.ndarray | list)
            norm_value = np.subtract(space_sample_arg, mu)
            norm_value = np.divide(norm_value, sigma)
            return norm_value.astype(np.float32)
        return copy.deepcopy(space_sample_arg)

    @staticmethod
    def unnormalize_sample_from_mu_sigma(
        space_likes: tuple[gymnasium.spaces.Space, OrderedDict | (dict | (tuple | (np.ndarray | list)))],
        mu: float = 0.0,
        sigma: float = 1,
    ) -> OrderedDict | (dict | (tuple | (np.ndarray | list))):
        """
        This unnormalizes a sample from a box space using the mu and sigma arguments.

        Parameters
        ----------
        space_likes: typing.Tuple[gymnasium.spaces.Space, sample_type]
            The first element is the gymnasium space.
            The second element is the sample of this space to scale.
        mu: float
            The mu used for un-normalizing the sample.
        sigma: float
            The sigma used for un-normalizing the sample.

        Returns
        -------
        typing.Union[OrderedDict, dict, tuple, np.ndarray, list]:
            The unnormalized sample.
        """
        (space_arg, space_sample_arg) = space_likes
        if isinstance(space_arg, gymnasium.spaces.Box):
            val = np.array(space_sample_arg)
            norm_value = np.add(np.multiply(val, sigma), mu)
            return norm_value.astype(np.float32)
        return copy.deepcopy(space_sample_arg)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
collection utils
"""

import collections
import typing


def get_dictionary_subset(data_dict: dict | typing.OrderedDict, keys: typing.Collection) -> typing.OrderedDict:
    """
    Create dictionary of input tuple with subset of values

    Parameters
    ----------
    data_dict typing.Union[typing.Dict, typing.OrderedDict]: src dictionary
    keys Collection: collection of keys defining subset

    Returns
    -------
    dictionary of input tuple with subset of values
    """
    return collections.OrderedDict({key: data_dict[key] for key in keys if key in data_dict})



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from collections import OrderedDict
from functools import cached_property

from corl.glues.base_glue import BaseAgentPlatformGlue, BaseAgentPlatformGlueValidator
from corl.libraries.property import DictProp, DiscreteProp
from corl.libraries.units import corl_quantity
from corl.simulators.base_parts import BasePlatformPart
from corl.simulators.common_platform_utils import get_part_by_name


class ObservePartValidityValidator(BaseAgentPlatformGlueValidator):
    """
    part: which part to find on the platform
    """

    part: str


class ObservePartValidity(BaseAgentPlatformGlue):
    """Glue to observe a part's validity flag"""

    class Fields:
        """
        Fields in this glue
        """

        VALIDITY_OBSERVATION = "validity_observation"

    @staticmethod
    def get_validator() -> type[ObservePartValidityValidator]:
        return ObservePartValidityValidator

    def __init__(self, **kwargs) -> None:
        self.config: ObservePartValidityValidator
        super().__init__(**kwargs)

        self._part: BasePlatformPart = get_part_by_name(self._platform, self.config.part)
        self._part_name: str = self.config.part

        self._uname = f"ObserveValidity_{self._part_name}"

    def get_unique_name(self) -> str:
        """Class method that retrieves the unique name for the glue instance"""
        return f"ObserveValidity_{self._part_name}"

    @cached_property
    def observation_prop(self):
        return DictProp(spaces={self.Fields.VALIDITY_OBSERVATION: DiscreteProp(n=2)})

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units):
        """Observation Values"""
        d = OrderedDict()
        d[self.Fields.VALIDITY_OBSERVATION] = corl_quantity()(int(self._part.valid), "dimensionless")
        return d



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from abc import ABC, abstractmethod

from corl.simulators.base_platform import BasePlatform


class PlatformSerializer(ABC):
    """Abstract class to define how to serialize a platform for saving"""

    @abstractmethod
    def serialize(self, platform: BasePlatform) -> dict:
        """Serialize provided platform

        Args:
            platform (BasePlatform): Platform to serialize

        Returns:
            dict: Platform serialized into a dict
        """
        raise RuntimeError("Must Instantiate this method")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Print a visualization to screen
"""
import os
import pathlib
import typing

import pandas as pd

# from dash import Dash, dcc, html, Input, Output
try:
    import plotly.express as px

    html_plots = True
except ImportError as e:
    print("HTMl Plots unavailable: ", e)
    html_plots = False

from corl.evaluation.metrics.types.nonterminals.timed_value import TimedValue
from corl.evaluation.metrics.types.nonterminals.vector import Vector

# from corl.evaluation.visualization.plot_configuration import PlotConfiguration, RangeMetric, RangePlot
from corl.evaluation.visualization.visualization import Visualization
from corl.libraries.units import Quantity

FUTURE = typing.Any


class HMTLPlots(Visualization):
    """Visualizer that prints to stdout"""

    def __init__(
        self,  # plot_configurations: typing.List[PlotConfiguration],
        *args,
        **kwargs,
    ):
        """Setup a Jupyter notebook visualization

        Parameters
        ----------
        done_means: str
            Path to checkpoint file
        done_variances: str
            File name that is being used during this evaluation
        """
        super().__init__(*args, **kwargs)
        # self._plot_configurations = plot_configurations

        if html_plots is False:
            raise RuntimeError("Can not utilize HTMLPlots with no plotly")

    def generate_plot(
        self,
        df,
    ) -> None:
        """Generate html plot using plotly"""

        filename = str(df.columns[0])

        # Figure out the title and directory_path for figure
        if len(df.columns) == 0:
            raise RuntimeError("Zero columns provided")
        if len(df.columns) == 1:
            title = filename
            participant_name, event_data, metric = filename.split("--")
            directory_path = os.path.join(self.out_folder, f"test_case_{event_data!s}", "plots", str(participant_name))
        else:
            all_names: list[str] = list(df.columns)

            # we now have a list of names, how should we combine?
            # We are going to assume the format is participant--test_case--metric
            # going to further assume that participant_name is the same and metric is the same
            participant_name, _, metric = all_names[0].split("--")

            # TODO: is something better here?
            #       Should we do some error handling to make sure everyones participant_name and metric are the same?

            title = str(participant_name + "--" + str(metric))
            directory_path = os.path.join(self.out_folder, "plots", str(participant_name))

        # Generate the figure
        fig = px.line(df, title=title)

        # Write figure to html
        pathlib.Path(directory_path).mkdir(parents=True, exist_ok=True)
        print("My current directory is : " + directory_path)
        fig.write_html(os.path.join(directory_path, str(metric) + ".html"))

    def __extract_evaluation_metric_as_dataframes(self) -> dict[str, FUTURE]:
        """This function creates a plot for each metric that contains all the runs, for each participant"""
        participant_dict: dict = {}
        metric_dict: dict = {}

        for participant in self._post_processed_data.participants:
            print(str(participant))
            for event in self._post_processed_data.participants[participant].events:
                # Search for metric on each event of the participant
                for metric, value in event.metrics.items():
                    # Can only plot a
                    if isinstance(value, Vector) and len(value.arr) > 0 and isinstance(value.arr[0], TimedValue):
                        arr: list[TimedValue] = value.arr  # type: ignore

                        if isinstance((arr[0].value.value), Quantity):  # type: ignore
                            data = list(zip([x.time.value for x in arr], [x.value.value.m for x in arr]))  # type: ignore
                        elif isinstance((value.arr[0].value.value[0]), float):  # type: ignore
                            data = list(zip([x.time.value for x in arr], [x.value.value[0] for x in arr]))  # type: ignore
                        else:
                            raise RuntimeError("Unknown how to extract")
                            # data = list(zip([x.time.value for x in value.arr], [x.value.value[0] for x in value.arr]))
                            # data = list(zip([x.time.value for x in value.arr], [x.value.value.value for x in value.arr]))

                        label = str(participant) + "--" + str(event.data.test_case) + "--" + str(metric)

                        df = pd.DataFrame(data, columns=["time", label])
                        df.set_index("time", inplace=True)
                        if metric not in metric_dict:
                            print("New metric found: " + str(metric))
                            metric_dict[metric] = []
                            metric_dict[metric].append(df)
                        else:
                            print("Metric " + str(metric) + " already exists. Append")
                            metric_dict[metric].append(df)

            # copy the dict so you won't lose data once its cleared
            complete_dict = metric_dict.copy()
            # clear metric_dict so it will be empty for the next participant
            metric_dict.clear()

            if str(participant) not in participant_dict:
                participant_dict[participant] = complete_dict
            else:
                participant_dict[participant].append(complete_dict)

        return participant_dict

    def visualize(self):  # I am okay with this method being long
        """Execute visualization"""

        # plot_list = self.__extract_event_metric_as_dataframes()
        # for range_df in plot_list:
        #     self.generate_plot(range_df, Filename.EVENT)

        # Generate dataframes for the plots we are going to generate
        plot_dict = self.__extract_evaluation_metric_as_dataframes()

        for plot_data in plot_dict.values():
            for metric in plot_data:
                # create aggregate dataframe to aggregate across test cases
                aggregate_df = pd.DataFrame(columns=["time"])
                aggregate_df.set_index("time", inplace=True)

                # Iterate over test cases
                for item in plot_data[metric]:
                    # generate individual test case plot
                    self.generate_plot(item)

                    # Add the data to the aggregate
                    aggregate_df = pd.concat([aggregate_df, item])

                # Generate the aggregate plot
                self.generate_plot(aggregate_df)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from collections import OrderedDict
from collections.abc import Callable

import numpy as np

ObsType = np.ndarray | tuple | dict


def mutate_observations(observations: OrderedDict, mutate_fn: Callable[[str, str, ObsType], ObsType]) -> OrderedDict:
    """
    Mutates observations according to mutate_fn
    if the MUTATED_OBSERVATION is not None, it will be included in the output

    Parameters
    ----------
    observations:
        An nested dictionary: observations[AGENT_ID][OBSERVATION_NAME] -> OBSERVATION
    mutate_fn(AGENT_ID: str, OBSERVATION_NAME: str, OBSERVATION: ObsType): MUTATED_OBSERVATION: ObsType
        Returns
        -------
            The mutated observation

    Returns
    -------
    OrderedDict:
        the mutated observation samples
    """
    mutated_observation_dict: OrderedDict = OrderedDict()

    for agent_id, obs_dict in observations.items():
        for obs_name, obs in obs_dict.items():
            mutated_obs = mutate_fn(agent_id, obs_name, obs)

            if mutated_obs is not None:
                if agent_id not in mutated_observation_dict:
                    mutated_observation_dict[agent_id] = OrderedDict()

                mutated_observation_dict[agent_id][obs_name] = mutated_obs

    return mutated_observation_dict


def filter_observations(observations: OrderedDict, filter_fn: Callable[[str, str, ObsType], bool]) -> OrderedDict:
    """
    Filters observations that don't match filter_fn

    Parameters
    ----------
    observations:
        An nested dictionary: observations[AGENT_ID][OBSERVATION_NAME] -> OBSERVATION
    filter_fn(AGENT_ID: str, OBSERVATION_NAME: str, OBSERVATION: ObsType): bool
        Returns
        -------
            true:  the observation will be included in the output
            false: the observation will not be included in the output

    Returns
    -------
    OrderedDict:
        the filtered observation samples
    """
    return mutate_observations(
        observations, lambda agent_id, obs_name, obs: obs if filter_fn(agent_id, obs_name, obs) else None  # type: ignore
    )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This base class (BasePlatformGlue) is the "glue" that sticks the base
integration object(s) to the environment.This class and its derived classes
are expected to be agnostic to the specific integration. The idea is that
you can then use the same "glue" for training and inference on any
integration.

The main purpose of the code in this class is to expose the specific actions
and observations for your given task. For example you may not want the full
controls from a BaseController for a given problem. This class is where you
decide to modify the controls exposed to the policy to only include what you
want actually controlled.

Deciding on how to combine actions is much more important when there are
multiple actions like with platform and sensor control. These connections also
apply to observations for instance with multiple observations like platform
state and sensor observations. How the actions and observations are "glued"
together is specific to the task and thus require an implementation of the base
class.

The glue you use must match the actual agent in the environment or be robust
enough to handle that agent. For example if there are different controllers
than expected do you throw an error or just reduce the action space.
A good practice is to assert the control or sensor is a type or a
subclass of a type to guarantee it has the functionality you expect.

The glue classes are modular in that you can use multiple glue classes per
agent. This allows stacking a glue to a stick controller glue to get
a and stick controlled agent. However, some control schemes may not be
compatible with each other. For example if you try to add a pitch control with
the stick and a pitch rate controller, the resulting behavior is not
straightforward and may cause your backend to throw an error.
"""
import abc
import enum
import logging
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium
import tree
from pydantic import BaseModel, ConfigDict, ImportString

from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.normalization import LinearNormalizer, Normalizer
from corl.libraries.property import Prop
from corl.libraries.units import Quantity
from corl.rewards.base_measurement_operation import ExtractorSet, ObservationExtractorValidator
from corl.simulators.base_platform import BasePlatform


class TrainingExportBehavior(str, enum.Enum):
    """Enumeration of training behaviors"""

    INCLUDE = "INCLUDE"
    EXCLUDE = "EXCLUDE"


class BaseAgentGlueNormalizationValidator(BaseModel):
    """
    enabled: if normalization is enabled for this glue
    minimum: the minimum value this glue's output will be normalized to
    maximum: the maximum value this glue's output will be normalized to
    """

    enabled: bool = True
    normalizer: ImportString = LinearNormalizer
    config: dict = {}


class BaseAgentGlueValidator(BaseModel):
    """
    name: the custom name of this glue
    agent_name: the name of the agent who owns this glue
    seed: temporary - assume this will be removed
    normalization: how to handle obs normalization, see BaseAgentGlueNormalizationValidator
    training_export_behavior: if this glues output should be included in the obs provided to policy
    clip_to_space: if this glues output should be clipped to match the obs space
    extractors: A dict of extractors to initialize for use with the topologically sorted
                    obs dictionary
    """

    name: str | None = None
    agent_name: str
    seed: int | None = None
    normalization: BaseAgentGlueNormalizationValidator = BaseAgentGlueNormalizationValidator()
    training_export_behavior: TrainingExportBehavior = TrainingExportBehavior.INCLUDE
    clip_to_space: bool = False
    extractors: dict[str, ObservationExtractorValidator] = {}


class BaseAgentGlue(abc.ABC):
    """
    BasePlatformGlue abstract class that provides the action space, observation space and how to apply actions and get observations
    for a platform

    """

    def __init__(self, **kwargs) -> None:
        """
        The init function for an Agent Glue class

        Parameters
        ----------
        state: typing.Tuple[BasePlatform]
            The initial state of the environment so the glue can stick together platform(s) or platform parts
        agent_id: str
            The name of the agent this glue is stuck to
        config: dict
            The configuration parameters of this glue class
        """
        self.config: BaseAgentGlueValidator = self.get_validator()(**kwargs)
        self._normalizer: Normalizer = self.config.normalization.normalizer(**self.config.normalization.config)
        self._agent_removed = False
        # construct extractors
        self.extractors: dict[str, ExtractorSet] = {}
        for key, extractor in self.config.extractors.items():
            self.extractors[key] = extractor.construct_extractors()

        self._normalized_action_space: gymnasium.Space[typing.Any] | None = None
        self._normalized_observation_space: gymnasium.Space[typing.Any] | None = None

    @staticmethod
    def get_validator() -> type[BaseAgentGlueValidator]:
        """returns the validator for this class

        Returns:
            BaseAgentGlueValidator -- A pydantic validator to be used to validate kwargs
        """
        return BaseAgentGlueValidator

    @abc.abstractmethod
    def get_unique_name(self) -> str:
        """Provides a unique name of the glue to differentiate it from other glues."""

    @cached_property
    def action_prop(self) -> Prop | None:
        """
        Build the action property for the controllers that defines the action this glue produces

        Returns
        -------
        typing.Optional[Prop]
            The Property that defines what this glue requires for an action
        """

    @cached_property
    def action_space(self) -> gymnasium.Space | None:
        """
        Build the action space for the controllers that defines the action given to apply_action
        This property should almost never need to be overridden

        Returns
        -------
        gymnasium.spaces.Space
            The gymnasium Space that defines the action given to the apply_action function
        """
        return self.action_prop.create_space() if self.action_prop else None

    @cached_property
    def normalized_action_space(self) -> gymnasium.Space | None:
        """
        Normalizes an action space from this glue to some normalized bounds.
        There is not rules on what "normal" is.
        The only idea is that the "normal" range is what the Policy will output

        The default implementation scales all Box spaces to a low=-1. and high =1.

        This function should be a dual function to the unnormalize_action

        Returns
        -------
        gymnasium.spaces.Space:
            The scaled action space
        """
        if not self._normalized_action_space and (action_space := self.action_space):
            self._normalized_action_space = (
                self._normalizer.normalize_space(space=action_space) if self.config.normalization.enabled else action_space
            )
        return self._normalized_action_space

    def apply_action(
        self,
        action: EnvSpaceUtil.sample_type,
        observation: EnvSpaceUtil.sample_type,
        action_space: gymnasium.Space,
        obs_space: gymnasium.Space,
        obs_units: OrderedDict,
    ) -> None:
        """
        Apply the action for the controllers

        Parameters
        ----------
        action
            The action for the class to apply to the platform
        observation
            The current observations before applying the action
        action_space: OrderedDict
            The action space for this agent
        obs_space: OrderedDict
            The observation space for this agent
        obs_units:
            The observation units for this agent
        """

    def unnormalize_action(self, action: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        Un-Normalizes an action space from this glue's normalized_action_space to some raw bounds

        The default implementation assumes the normalized scale was low=-1. and high =1.

        This function should be a dual function to the normalize_action_space

        Parameters
        ----------
        action: EnvSpaceUtil.sample_type
            The action to unnormalize
        Returns
        -------
        EnvSpaceUtil.sample_type:
            the unnormalized action
        """
        ret: EnvSpaceUtil.sample_type = action
        if self.config.normalization.enabled:
            ret = self._normalizer.unnormalize_sample(space=self.action_space, sample=action)  # type: ignore
        return ret

    @cached_property
    @abc.abstractmethod
    def observation_prop(self) -> Prop:
        """
        Build the property descriving the observation space for
        the platform using the state of the platform, controller, sensors, etc.

        Returns
        -------
        Prop
            The corl Prop that defines the returned space from the get_observation function
        """

    @cached_property
    def observation_space(self) -> gymnasium.Space | None:
        """
        Build the observation space for the platform using the state of the platform, controller, sensors, etc.
        This property should almost never need to be overridden

        Returns
        -------
        gymnasium.spaces.Space
            The gymnasium space that defines the returned space from the get_observation function
        """
        return self.observation_prop.create_space() if self.observation_prop else None

    @cached_property
    def normalized_observation_space(self) -> gymnasium.Space | None:
        """
        Normalizes an observation space from this glue to some normalized bounds.
        There is not rules on what "normal" is.
        The only idea is that the "normal" range is what the Policy will take as input for observations

        The default implementation scales all Box spaces to a low=-1. and high =1.

        This function should be the same as the normalize_observation function but return a Space and not the sample

        Returns
        -------
        gymnasium.spaces.Space:
            The scaled observation space
        """
        if not self._normalized_observation_space and (observation_space := self.observation_space):
            if self.config.normalization.enabled:
                tmp = self._normalizer.normalize_space(space=observation_space)
                # if not isinstance(tmp, gymnasium.spaces.Dict):
                #     raise RuntimeError(
                #         f"{self.get_unique_name()} tried to return something other than a dict observation_space, "
                #         "this is currently not supported."
                #     )
                self._normalized_observation_space = tmp
            else:
                self._normalized_observation_space = observation_space
        return self._normalized_observation_space

    @abc.abstractmethod
    def get_observation(self, other_obs: OrderedDict, obs_space: gymnasium.Space, obs_units: OrderedDict) -> EnvSpaceUtil.sample_type:
        """
        Get the actual observation for the platform using the state of the platform, controller, sensors, etc.

        Parameters
        ----------
        other_obs: OrderedDict
            The observation dict containing any observations this glue may depend on
        obs_space: OrderedDict
            The observation space for this agent
        obs_units:
            The obserbation units for this agent

        Returns
        -------
        EnvSpaceUtil.sample_type
            The actual observation for this platform from this glue class
        """

    def normalize_observation(self, observation: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        Normalizes an observation from this glue to some normalized bounds.
        There is not rules on what "normal" is.
        The only idea is that the "normal" range is what the Policy will use as observations

        Parameters
        ----------
        observation: EnvSpaceUtil.sample_type
            The observation we want to scale
        Returns
        -------
        EnvSpaceUtil.sample_type:
            The scaled observation
        """
        ret = tree.map_structure(lambda v: v.m if isinstance(v, Quantity) else v, observation)
        if self.config.normalization.enabled:
            obs_space = self.observation_space
            # if isinstance(obs_space, gymnasium.spaces.Dict) and any(isinstance(obs_space[key], Repeated) for key in obs_space.keys()):
            #     ret = self._normalizer.normalize_sample(space=obs_space, sample=copy.deepcopy(ret))
            # elif obs_space:
            ret = self._normalizer.normalize_sample(space=obs_space, sample=ret)  # type: ignore
        return ret

    def get_info_dict(self) -> EnvSpaceUtil.sample_type:  # noqa: PLR6301
        """
        Get the user specified metadata/metrics/etc.

        Returns
        -------
        EnvSpaceUtil.sample_type
            The actual info dict object for this platform from this glue class
        """
        return {}

    def agent_removed(self) -> bool:
        """
        Returns true if the agent has been removed, false otherwise
        """
        return self._agent_removed

    def set_agent_removed(self, agent_removed: bool = True) -> None:
        """
        Notify the glue that the agent it is 'attached' to has been removed by the simulation
        """
        self._agent_removed = agent_removed


class BaseAgentPlatformGlueValidator(BaseAgentGlueValidator):
    """
    platform: The platform object that this glue will read from/apply actions to
    """

    platform: BasePlatform
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseAgentPlatformGlue(BaseAgentGlue, abc.ABC):
    """
    BaseAgentPlatformGlue assumes that this glue corresponds to exactly one platform with the same
    platform name as the agent_id this glue is attached to

    This will then set the self._platform to this platform
    """

    def __init__(self, **kwargs) -> None:
        self.config: BaseAgentPlatformGlueValidator
        super().__init__(**kwargs)
        self._platform: BasePlatform = self.config.platform
        self._logger = logging.getLogger(BaseAgentPlatformGlue.__name__)

    @staticmethod
    def get_validator() -> type[BaseAgentPlatformGlueValidator]:
        return BaseAgentPlatformGlueValidator


class BaseAgentControllerGlue(BaseAgentPlatformGlue, abc.ABC):
    """
    BaseAgentControllerGlue assumes that this glue is for a controller and that an associated 'get_applied_control' method is available
    """

    @abc.abstractmethod
    def get_applied_control(self) -> Quantity:
        """
        Get the currently applied controls

        Returns
        -------
        Quantity
            The currently applied control
        """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import itertools
import typing
from collections import Counter

from corl.evaluation.metrics.generator import MetricGeneratorAggregator
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.nonterminals.dict import Dict
from corl.evaluation.metrics.types.nonterminals.vector import Vector
from corl.evaluation.metrics.types.terminals.discrete import Discrete
from corl.evaluation.metrics.types.terminals.rate import Rate
from corl.evaluation.metrics.types.terminals.real import Real
from corl.evaluation.metrics.types.terminals.string import String


class TerminalCounterMetric(MetricGeneratorAggregator):
    """Count the instances of each value, returning count or percentage"""

    divide_total: bool = False

    def generate_metric(self, params: list[Metric] | Metric, **kwargs) -> Metric:
        """Generate the metric

        Arguments:
            params {typing.Union[typing.List[Metric], Metric]} -- Already computed metrics

        Returns:
            Metric -- Computed Metric
        """

        count: typing.Counter[typing.Any] = Counter()
        arr: list[Metric] = [params] if isinstance(params, Metric) else params
        for elem in arr:
            count.update(self._get_values(elem))

        if self.divide_total:
            return Dict({k: Real(v / len(arr)) for k, v in count.items()})

        return Dict({k: Rate(v, len(arr)) for k, v in count.items()})

    @classmethod
    def _get_values(cls, metric_elem: Metric) -> typing.Iterable[typing.Any]:
        if isinstance(metric_elem, Real | String | Discrete):
            return [metric_elem.value]

        if isinstance(metric_elem, Rate):
            return [metric_elem.occurrence / metric_elem.total]

        if isinstance(metric_elem, Vector):
            return itertools.chain.from_iterable(cls._get_values(x) for x in metric_elem.arr)

        raise TypeError(f"Unsupported metric: {type(metric_elem).__name__}")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.generator import MetricGeneratorTerminalEventScope
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.nonterminals.dict import Dict
from corl.evaluation.metrics.types.nonterminals.vector import Vector
from corl.evaluation.metrics.types.terminals.real import Real
from corl.evaluation.metrics.types.terminals.void import Void


class TotalReward(MetricGeneratorTerminalEventScope):
    """Generates single Real indicating the total reward for an event

    Metric: Real
    Scope: Event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:  # noqa: PLR6301
        if "agent_id" not in kwargs:
            raise RuntimeError('Expecting "agent_id" to be provided')

        agent_id = kwargs["agent_id"].split(".")[0]

        # Find the last step with observation data from the platform in question
        # If the platform isn't in the episode at all, return Void
        steps_with_platform_in_question = [item for item in params.steps if agent_id in item.agents and item.agents[agent_id] is not None]
        if not steps_with_platform_in_question:
            return Void()

        last_step_with_platform_data = steps_with_platform_in_question[-1]

        if last_step_with_platform_data.agents[agent_id] is None:
            raise RuntimeError("Non Op")

        return Real(last_step_with_platform_data.agents[agent_id].total_reward)


class RewardVector(MetricGeneratorTerminalEventScope):
    """Generates vector of Dicts for the rewards calculated for each step during an event for an agent

    Metric: Vector[Dict]
    Scope: Event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:  # noqa: PLR6301
        if "agent_id" not in kwargs:
            raise RuntimeError('Expecting "agent_id" to be provided')

        agent_id = kwargs["agent_id"].split(".")[0]

        arr: list[Metric] = []
        for step in params.steps:
            if agent_id not in step.agents or step.agents[agent_id] is None:
                break

            map_rewards = step.agents[agent_id].rewards

            if map_rewards is None:
                continue
            # Create a non terminal metric (Dict) that is comprised of the terminal (Real) rewards
            real_dict: dict[str, Metric] = {key: Real(map_rewards[key]) for key in map_rewards}
            arr.append(Dict(real_dict))

        return Vector(arr)



"""
This module contains implementations of Sensors that reside on the Docking1dPlatform.
"""
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_parts import BaseSensor
from corl.simulators.docking_1d.available_platforms import Docking1dAvailablePlatformType
from corl.simulators.docking_1d.properties import PositionProp, VelocityProp
from corl.simulators.docking_1d.simulator import Docking1dSimulator


class PositionSensor(BaseSensor):
    """
    A Sensor to measure the position of the associated Docking1dPlatform.
    """

    def __init__(self, parent_platform, config, measurement_properties=PositionProp):
        super().__init__(parent_platform=parent_platform, config=config, property_class=measurement_properties)

    def _calculate_measurement(self, state) -> Quantity:
        """
        get measurements from the sensor
        """
        return corl_get_ureg().Quantity(self.parent_platform.position.astype("float32"), "meter")


class VelocitySensor(BaseSensor):
    """
    A Sensor to measure the velocity of the associated Docking1dPlatform.
    """

    def __init__(self, parent_platform, config, measurement_properties=VelocityProp):
        super().__init__(parent_platform=parent_platform, config=config, property_class=measurement_properties)

    def _calculate_measurement(self, state) -> Quantity:
        """
        get measurements from the sensor
        """
        return corl_get_ureg().Quantity(self.parent_platform.velocity.astype("float32"), "meter / second")


# Register sensors with PluginLibrary. Requires a class, reference name, and a dict of Simulator class and platform
# type enum.

PluginLibrary.AddClassToGroup(
    PositionSensor, "1D_Sensor_Position", {"simulator": Docking1dSimulator, "platform_type": Docking1dAvailablePlatformType}
)

PluginLibrary.AddClassToGroup(
    VelocitySensor, "1D_Sensor_Velocity", {"simulator": Docking1dSimulator, "platform_type": Docking1dAvailablePlatformType}
)



"""
This module contains implementations of Sensors that reside on the Docking1dPlatform.
"""
from typing import Annotated

import numpy as np
from pydantic import Field, StrictFloat

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import BoxProp, DiscreteProp
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_parts import BasePlatformPartValidator, BaseSensor
from corl.simulators.pong.available_platforms import PongAvailablePlatformType
from corl.simulators.pong.paddle_platform import PaddleType
from corl.simulators.pong.simulator import PongSimulator


class PositionProp(BoxProp):
    """
    Position sensor properties.

    name : str
        sensor property name
    low : list[float]
        minimum bounds of sensor output
    high : list[float]
        maximum bounds of sensor output
    unit : str
        unit of measurement for sensor output
    description : str
        description of sensor properties
    """

    name: str = "position"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [1000.0]
    unit: str = "dimensionless"
    description: str = "Position Sensor Properties"


class SizeProp(BoxProp):
    """
    paddle size sensor properties.

    name : str
        sensor property name
    low : list[float]
        minimum bounds of sensor output
    high : list[float]
        maximum bounds of sensor output
    unit : str
        unit of measurement for sensor output
    description : str
        description of sensor properties
    """

    name: str = "paddle_size"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [1000.0]
    unit: str = "dimensionless"
    description: str = "paddle Size Sensor Properties"


class PosVelProp(BoxProp):
    """
    Position sensor properties.

    name : str
        sensor property name
    low : list[float]
        minimum bounds of sensor output
    high : list[float]
        maximum bounds of sensor output
    unit : str
        unit of measurement for sensor output
    description : str
        description of sensor properties
    """

    name: str = "ball_state"
    low: Annotated[list[StrictFloat], Field(min_length=4, max_length=4)] = [-50.0, -50.0, -100.0, -100.0]
    high: Annotated[list[StrictFloat], Field(min_length=4, max_length=4)] = [1000.0, 1000.0, 100.0, 100.0]
    unit: str = "dimensionless"
    description: str = "Position and Velocity Sensor Properties"


class PlatformSideProp(DiscreteProp):
    """PlatformSideProp"""

    name: str = "side"
    n: int = 2
    unit: str = "dimensionless"
    description: str = "Side of the platform"


class PaddlePositionSensorValidator(BasePlatformPartValidator):
    """PaddlePositionSensorValidator"""

    paddle: PaddleType


class PaddlePositionSensor(BaseSensor):
    """
    A Sensor to measure the position of paddles.
    """

    def __init__(self, parent_platform, config, measurement_properties=PositionProp) -> None:
        self.config: PaddlePositionSensorValidator
        super().__init__(parent_platform=parent_platform, config=config, property_class=measurement_properties)

    @staticmethod
    def get_validator() -> type[PaddlePositionSensorValidator]:
        """
        Get the validator used to validate the kwargs passed to BaseAgent.

        Returns
        -------
        BaseAgentParser
            A BaseAgent kwargs parser and validator.
        """
        return PaddlePositionSensorValidator

    def _calculate_measurement(self, state) -> Quantity:
        """get paddle y position

        Parameters
        ----------
        state: BaseSimulatorState
            current simulations state

        Returns
        -------
        y: np.array
            current y position of paddle
        """
        val = state.pong_game.right_paddle.y
        if self.config.paddle is PaddleType.LEFT:
            val = state.pong_game.left_paddle.y
        return corl_get_ureg().Quantity(np.array([val], dtype=np.float32), self._properties.unit)


class PaddleSizeSensorValidator(BasePlatformPartValidator):
    """PaddleSizeSensorValidator"""

    paddle: PaddleType


class PaddleSizeSensor(BaseSensor):
    """
    A Sensor to measure the size of paddles.
    """

    def __init__(self, parent_platform, config, measurement_properties=SizeProp):
        self.config: PaddleSizeSensorValidator  # type: ignore
        super().__init__(parent_platform=parent_platform, config=config, property_class=measurement_properties)

    @staticmethod
    def get_validator() -> type[PaddleSizeSensorValidator]:
        """
        Get the validator used to validate the kwargs passed to BaseAgent.

        Returns
        -------
        BaseAgentParser
            A BaseAgent kwargs parser and validator.
        """
        return PaddleSizeSensorValidator

    def _calculate_measurement(self, state) -> Quantity:
        """get paddle y size

        Parameters
        ----------
        state: BaseSimulatorState
            current simulations state

        Returns
        -------
        y: np.array
            y size of paddle
        """
        val = state.pong_game.right_paddle.height
        if self.config.paddle is PaddleType.LEFT:
            val = state.pong_game.left_paddle.height
        return corl_get_ureg().Quantity(np.array([val], dtype=np.float32), self._properties.unit)


class BallSensor(BaseSensor):
    """
    A Sensor to measure the velocity of the associated Docking1dPlatform.
    """

    def __init__(self, parent_platform, config, measurement_properties=PosVelProp):
        super().__init__(parent_platform=parent_platform, config=config, property_class=measurement_properties)

    def _calculate_measurement(self, state) -> Quantity:  # noqa: PLR6301
        """get paddle y position

        Parameters
        ----------
        state: BaseSimulatorState
            current simulations state

        Returns
        -------
        ball_state: np.array
            four element array contain x, y position and velocity of the ball
        """
        return corl_get_ureg().Quantity(
            np.array(
                [state.pong_game.ball.x, state.pong_game.ball.y, state.pong_game.ball.x_vel, state.pong_game.ball.y_vel], dtype=np.float32
            ),
            "dimensionless",
        )


class PlatformPaddleTypeSensor(BaseSensor):
    """
    A Sensor to the provide type of a paddle (left or right).
    """

    def __init__(self, parent_platform, config, measurement_properties=PlatformSideProp):
        super().__init__(parent_platform=parent_platform, config=config, property_class=measurement_properties)

    def _calculate_measurement(self, state) -> Quantity:
        """get paddle side

        Parameters
        ----------
        state: BaseSimulatorState
            current simulations state

        Returns
        -------
        y: int
            current pade side
        """
        val = 1
        if self.parent_platform.paddle_type is PaddleType.LEFT:
            val = 0
        return corl_get_ureg().Quantity(np.array(val, dtype=np.int32), self._properties.unit)


# Register sensors with PluginLibrary. Requires a class, reference name, and a dict of Simulator class and platform
# type enum.

PluginLibrary.AddClassToGroup(
    PlatformPaddleTypeSensor, "Platform_Paddle_Type_Sensor", {"simulator": PongSimulator, "platform_type": PongAvailablePlatformType}
)

PluginLibrary.AddClassToGroup(
    PaddlePositionSensor, "Paddle_Position_Sensor", {"simulator": PongSimulator, "platform_type": PongAvailablePlatformType}
)

PluginLibrary.AddClassToGroup(
    PaddleSizeSensor, "Paddle_Size_Sensor", {"simulator": PongSimulator, "platform_type": PongAvailablePlatformType}
)

PluginLibrary.AddClassToGroup(BallSensor, "Ball_Sensor", {"simulator": PongSimulator, "platform_type": PongAvailablePlatformType})



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Base Integration Controls Module
"""
import abc

from corl.simulators.base_parts import BaseController


class BaseRollPitchYawSpeedController(BaseController, abc.ABC):
    """
    BaseRollPitchYawSpeedController abstraction of controlling an aircraft
    """


class BaseSpeedController(BaseController, abc.ABC):
    """[summary]

    Arguments:
        BaseController: [description]
        abc: [description]

    Returns:
        [type] -- [description]
    """


class BaseRollController(BaseController, abc.ABC):
    """[summary]

    Arguments:
        BaseController: [description]
        abc: [description]

    Returns:
        [type] -- [description]
    """


class BasePitchController(BaseController, abc.ABC):
    """[summary]

    Arguments:
        BaseController: [description]
        abc: [description]

    Returns:
        [type] -- [description]
    """


class BaseYawController(BaseController, abc.ABC):
    """[summary]

    Arguments:
        BaseController: [description]
        abc: [description]

    Returns:
        [type] -- [description]
    """


class BasePitchRollController(BaseController, abc.ABC):
    """[summary]

    Arguments:
        BaseController: [description]
        abc: [description]

    Returns:
        [type] -- [description]
    """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import typing

from pydantic import BaseModel, PrivateAttr

from corl.libraries.factory import Factory


class Condition(BaseModel):
    """Creates boolean condition logic from a lhs and a operator

    Results in a boolean function which the lhs is to be provided
    """

    operator: str
    lhs: int | (float | dict)

    _func: typing.Callable[[typing.Any], bool] = PrivateAttr()
    """
    This lambda is private, it is built in the constructor based on non private members
    """

    def __init__(self, **data):
        super().__init__(**data)

        # build the lhs
        if isinstance(self.lhs, float | int):
            lhs = self.lhs
        elif isinstance(self.lhs, dict):
            if "functor" in self.lhs:
                lhs = Factory(type=self.lhs["functor"], config=self.lhs["config"]).build()
            else:
                raise RuntimeError("Unknown how to build lhs")
        else:
            raise RuntimeError(f"Unknown type {type(self.lhs)} on the lhs of CriteriaRate")

        # Construct the criteria_func method
        if self.operator == "==":
            self._func = lambda x: x == lhs
        elif self.operator == "<":
            self._func = lambda x: x < lhs
        elif self.operator == ">":
            self._func = lambda x: x > lhs
        else:
            raise RuntimeError(f"Unknown Operator: {self.operator}")

    def func(self, rhs: typing.Any) -> bool:
        """Execute the function wrapped up in this conditional

        Args:
            rhs (typing.Any): Right hand side

        Returns:
            bool: Result of condition
        """
        return self._func(rhs)


class NamedCondition(BaseModel):
    """Wraps a Condition with a name"""

    type: str  # noqa: A003
    condition: Condition



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from typing import Annotated

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field

from corl.evaluation.default_config_updates import DoNothingConfigUpdate
from corl.evaluation.evaluation_factory import FactoryLoader

from .config_updater import ConfigUpdate
from .environment_state_extractor import EnvironmentStateExtractor
from .platform_serializer import PlatformSerializer


class Plugins(BaseModel):
    """Plugins that define environment specific information to run an evaluation"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    platform_serialization: Annotated[PlatformSerializer, BeforeValidator(FactoryLoader.resolve_factory)]
    eval_config_update: list[Annotated[ConfigUpdate, BeforeValidator(FactoryLoader.resolve_factory)]] = Field(
        validate_default=True, default_factory=lambda: [DoNothingConfigUpdate()]
    )

    environment_state_extractor: Annotated[EnvironmentStateExtractor, BeforeValidator(FactoryLoader.resolve_factory)] | None = None



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Reward Functor Base Module
"""
import abc
from collections import OrderedDict

import gymnasium
from pydantic import BaseModel

from corl.libraries.env_func_base import EnvFuncBase
from corl.simulators.base_simulator_state import BaseSimulatorState


class RewardFuncBaseValidator(BaseModel):
    """
    name: Name of reward functor
    agent_name: Name of agent the reward functor belongs to
    platform_names: list of platforms controlled by the agent
                    this reward belongs to
    """

    name: str | None = None
    agent_name: str
    platform_names: list[str]


class RewardFuncBase(EnvFuncBase):
    """The base implementation for reward functors"""

    def __init__(self, **kwargs) -> None:
        self.config: RewardFuncBaseValidator = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[RewardFuncBaseValidator]:
        """Returns pydantic validator associated with this class"""
        return RewardFuncBaseValidator

    @abc.abstractmethod
    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: BaseSimulatorState,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> float:
        ...

    def post_process_trajectory(self, agent_id, state, batch, episode, policy):
        """Allows the user to modify the trajectory of the episode
        in the batch collected during an rllib callback. WARNING: This function is dangerous
        you can completely destroy training using this
        Use it only as a last resort
        """

    @property
    def name(self) -> str:
        """gets the name for the functor

        Returns
        -------
        str
            The name of the functor
        """
        return type(self).__name__ if self.config.name is None else self.config.name



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from collections.abc import Mapping

from corl.episode_parameter_providers import EpisodeParameterProviderValidator
from corl.episode_parameter_providers.base_network_epp import BaseNetworkParameterProvider, NetworkedParameterProviderValidator
from corl.evaluation.connection.base_epp_update import get_updates
from corl.evaluation.runners.inference.protocol.simple_epp_update import SimpleEppUpdate
from corl.libraries.parameters import OverridableParameterWrapper, ParameterValidator
from corl.libraries.units import Quantity, corl_quantity


class SimpleNetworkEppValidator(NetworkedParameterProviderValidator[SimpleEppUpdate]):
    @property
    def prefix(self) -> str:
        if self.epp_variety == "environment":
            return "environment"
        if not self.agent_id:
            raise RuntimeError("Sanity Check Fail - 'agent_id' should be valid")
        return self.agent_id


class SimpleNetworkEpp(BaseNetworkParameterProvider[SimpleEppUpdate]):
    def __init__(self, **kwargs):
        self.config: SimpleNetworkEppValidator  # type: ignore
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[EpisodeParameterProviderValidator]:
        return SimpleNetworkEppValidator

    def _do_update_params(
        self, params: Mapping[tuple[str, ...], OverridableParameterWrapper], updates: SimpleEppUpdate
    ) -> Mapping[tuple[str, ...], OverridableParameterWrapper]:
        update_dict = get_updates(updates, self.config.prefix)

        def try_get_units(param: OverridableParameterWrapper) -> str:
            if isinstance(param.config, ParameterValidator):
                return param.config.units
            return ""

        for key, value in update_dict.items():
            if key not in params:
                raise RuntimeError(f"Parameter '{key}' not found in params")

            if isinstance(value, str):
                params[key].override_value = value
            else:
                param_units = try_get_units(params[key])
                if isinstance(value, Quantity):
                    params[key].override_value = value.to(param_units)
                else:
                    params[key].override_value = corl_quantity()(value=value, units=param_units)

        return params



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This module extends corl.simulators.base_platform.BasePlatform to create a paddle platform
used to interact with the pong
"""

from enum import Enum

import pygame

from corl.simulators.base_platform import BasePlatform, BasePlatformValidator


class PaddleType(str, Enum):
    """
    Enumeration describing the paddle side [left or right]
    """

    LEFT = "left"
    RIGHT = "right"


class PaddlePlatformValidator(BasePlatformValidator):
    """paddle platform validator"""

    paddle_type: PaddleType


class PaddlePlatform(BasePlatform):
    """
    A platform representing a pong paddles.
    Allows for saving of "moves" to give an action to the pong game
    during the environment step function

    Parameters
    ----------
    platform_name : str
        Name of the platform
    platform : PaddleType
        The paddle side corresponding to what the platform in controlling
        in the pong game
    platform_config : dict
        Platform-specific configuration dictionary
    paddle: paddle
        The specific paddle object from the simulator
    """

    def __init__(self, platform_name, parts_list, paddle_type, paddle) -> None:
        self.config: PaddlePlatformValidator
        super().__init__(
            platform_name=platform_name,
            parts_list=parts_list,
            paddle_type=paddle_type,
        )
        self._last_move = pygame.K_0
        self.paddle = paddle

    @staticmethod
    def get_validator() -> type[PaddlePlatformValidator]:
        """Returns the paddle type associated with this platform

        Returns
        -------
        PaddleType
            the paddle type associated with this platform
        """
        return PaddlePlatformValidator

    @property
    def paddle_type(self) -> PaddleType:
        """Returns the paddle type associated with this platform

        Returns
        -------
        PaddleType
            the paddle type associated with this platform
        """
        return self.config.paddle_type

    @property
    def last_move(self):
        """
        return last action in units used by pong game env
        """
        return self._last_move

    @last_move.setter
    def last_move(self, value):
        """
        set last action in units used by pong game env
        """
        self._last_move = value

    @property
    def operable(self):
        """a paddle without collision is useless so we mark it inoperable
        allows us to test inoperable code paths in commander pong
        """
        return self.paddle.collision_on



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from corl.dones.done_func_base import DoneStatusCodes
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.generator import MetricGeneratorTerminalEventScope
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.nonterminals.vector import Vector
from corl.evaluation.metrics.types.terminals.discrete import Discrete
from corl.evaluation.metrics.types.terminals.string import String


# temporarily setting line 37 to 'DockingDoneStatus'
class StatusCode(MetricGeneratorTerminalEventScope):
    """Generates DoneStatusCodes value for the agent in an event

    Metric: Discrete
    Scope: Event
    """

    done_condition: str

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:
        if "agent_id" not in kwargs:
            raise RuntimeError('Expecting "agent_id" to be provided')

        agent_id = kwargs["agent_id"].split(".")[0]

        return Discrete[DoneStatusCodes](params.episode_state[agent_id][self.done_condition])


class DonesVec(MetricGeneratorTerminalEventScope):
    """Generates Vector of strings indicating the done condition

    Metric: Vector[String]
    Scope: Event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:  # noqa: PLR6301
        if "agent_id" not in kwargs:
            raise RuntimeError('Expecting "agent_id" to be provided')

        agent_id = kwargs["agent_id"]

        platform_id_list = params.agent_to_platforms[agent_id]

        done_set = set()

        for platform_id in platform_id_list:
            dones = params.dones[platform_id]
            triggered_dones = [key for key in dones if dones[key] is True]
            done_set.update(triggered_dones)

        return Vector([String(item) for item in done_set])
